from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Annotated, Any, NewType, dataclass_transform, get_origin

from sonolus.backend.mode import Mode
from sonolus.backend.ops import Op
from sonolus.script.internal.context import ctx
from sonolus.script.internal.impl import meta_fn
from sonolus.script.internal.introspection import get_field_specifiers
from sonolus.script.internal.native import native_function
from sonolus.script.interval import Interval
from sonolus.script.pointer import _deref
from sonolus.script.record import Record
from sonolus.script.sprite import Sprite


class JudgmentWindow(Record):
    """The window for judging the accuracy of a hit.

    Usage:
        ```python
        JudgmentWindow(perfect: Interval, great: Interval, good: Interval)
        ```
    """

    perfect: Interval
    """Interval for a perfect hit."""

    great: Interval
    """Interval for a great hit."""

    good: Interval
    """Interval for a good hit."""

    def update(
        self,
        perfect: Interval | None = None,
        great: Interval | None = None,
        good: Interval | None = None,
    ):
        """Update the window with the given intervals.

        Args:
            perfect: The interval for a perfect hit.
            great: The interval for a great hit.
            good: The interval for a good hit.
        """
        if perfect is not None:
            self.perfect = perfect
        if great is not None:
            self.great = great
        if good is not None:
            self.good = good

    def judge(self, actual: float, target: float) -> Judgment:
        """Judge the accuracy of a hit.

        Args:
            actual: The actual time of the hit.
            target: The target time of the hit.

        Returns:
            The [`Judgment`][sonolus.script.bucket.Judgment] of the hit.
        """
        return _judge(
            actual,
            target,
            *self.perfect.tuple,
            *self.great.tuple,
            *self.good.tuple,
        )

    def __mul__(self, other: float | int) -> JudgmentWindow:
        """Multiply the intervals by a scalar."""
        return JudgmentWindow(
            self.perfect * other,
            self.great * other,
            self.good * other,
        )

    def __add__(self, other: float | int) -> JudgmentWindow:
        """Add a scalar to the intervals."""
        return JudgmentWindow(
            self.perfect + other,
            self.great + other,
            self.good + other,
        )

    @property
    def start(self) -> float:
        """The start time of the good interval."""
        return self.good.start

    @property
    def end(self) -> float:
        """The end time of the good interval."""
        return self.good.end


class Judgment(IntEnum):
    """The judgment of a hit."""

    MISS = 0
    PERFECT = 1
    GREAT = 2
    GOOD = 3


@native_function(Op.Judge)
def _judge(
    actual: float,
    target: float,
    perfect_min: float,
    perfect_max: float,
    great_min: float,
    great_max: float,
    good_min: float,
    good_max: float,
) -> Judgment:
    diff = actual - target
    if perfect_min <= diff <= perfect_max:
        return Judgment.PERFECT
    if great_min <= diff <= great_max:
        return Judgment.GREAT
    if good_min <= diff <= good_max:
        return Judgment.GOOD
    return Judgment.MISS


class Bucket(Record):
    """A bucket for entity judgment results.

    Usage:
        ```python
        Bucket(id: int)
        ```
    """

    id: int
    """Bucket ID."""

    @property
    @meta_fn
    def window(self) -> JudgmentWindow:
        """The judgment window of the bucket."""
        if not ctx():
            raise RuntimeError("Bucket window access outside of compilation")
        match ctx().global_state.mode:
            case Mode.PLAY:
                return _deref(ctx().blocks.LevelBucket, self.id * JudgmentWindow._size_(), JudgmentWindow)
            case Mode.WATCH:
                return _deref(ctx().blocks.LevelBucket, self.id * JudgmentWindow._size_(), JudgmentWindow)
            case _:
                raise RuntimeError("Invalid mode for bucket window access")

    @window.setter
    @meta_fn
    def window(self, value: JudgmentWindow):
        if not ctx():
            raise RuntimeError("Bucket window access outside of compilation")
        self.window.update(value.perfect, value.great, value.good)


@dataclass
class _BucketSprite:
    id: int
    fallback_id: int | None
    x: float
    y: float
    w: float
    h: float
    rotation: float

    def to_dict(self):
        results = {
            "id": self.id,
            "x": self.x,
            "y": self.y,
            "w": self.w,
            "h": self.h,
            "rotation": self.rotation,
        }
        if self.fallback_id is not None:
            results["fallbackId"] = self.fallback_id
        return results


@dataclass
class _BucketInfo:
    sprites: list[_BucketSprite]
    unit: str | None = None

    def to_dict(self):
        results = {
            "sprites": [sprite.to_dict() for sprite in self.sprites],
        }
        if self.unit is not None:
            results["unit"] = self.unit
        return results


def bucket_sprite(
    *,
    sprite: Sprite,
    fallback_sprite: Sprite | None = None,
    x: float,
    y: float,
    w: float,
    h: float,
    rotation: float = 0,
) -> _BucketSprite:
    """Define a sprite for a bucket."""
    return _BucketSprite(sprite.id, fallback_sprite.id if fallback_sprite else None, x, y, w, h, rotation)


def bucket(*, sprites: list[_BucketSprite], unit: str | None = None) -> Any:
    """Define a bucket with the given sprites and unit."""
    return _BucketInfo(sprites, unit)


type Buckets = NewType("Buckets", Any)


@dataclass_transform()
def buckets[T](cls: type[T]) -> T | Buckets:
    """Decorator to define a buckets class.

    Usage:
        ```python
        @buckets
        class Buckets:
            note: Bucket = bucket(
                sprites=[
                    bucket_sprite(
                        sprite=Skin.note,
                        x=0,
                        y=0,
                        w=2,
                        h=2,
                    )
                ],
                unit=StandardText.MILLISECOND_UNIT,
            )
        ```
    """
    if len(cls.__bases__) != 1:
        raise ValueError("Buckets class must not inherit from any class (except object)")
    instance = cls()
    bucket_info = []
    for i, (name, annotation) in enumerate(get_field_specifiers(cls).items()):
        if get_origin(annotation) is not Annotated:
            raise TypeError(f"Invalid annotation for buckets: {annotation}")
        annotation_type = annotation.__args__[0]
        annotation_values = annotation.__metadata__
        if annotation_type is not Bucket:
            raise TypeError(f"Invalid annotation for buckets: {annotation}, expected annotation of type Bucket")
        if len(annotation_values) != 1 or not isinstance(annotation_values[0], _BucketInfo):
            raise TypeError(
                f"Invalid annotation for buckets: {annotation}, expected a single BucketInfo annotation value"
            )
        info = annotation_values[0]
        bucket_info.append(info)
        setattr(instance, name, Bucket(i))
    instance._buckets_ = bucket_info
    instance._is_comptime_value_ = True
    return instance


@buckets
class EmptyBuckets:
    pass



import json
from pathlib import Path
from typing import TextIO

functions_file = "Functions.json"
block_files = {
    "Tutorial": "Engine/Tutorial/Blocks.json",
    "Play": "Level/Play/Blocks.json",
    "Preview": "Level/Preview/Blocks.json",
    "Watch": "Level/Watch/Blocks.json",
}
base_dir = Path(__file__).parent
out_dir = base_dir / "out"
runtimes_dir = base_dir / "runtimes"


def functions():
    with (
        (runtimes_dir / functions_file).open("r", encoding="utf-8") as f,
        (out_dir / "ops.py").open("w", encoding="utf-8") as out,
    ):
        data = json.load(f)
        out.write(
            "from enum import Enum\n"
            "\n"
            "\n"
            "class Op(str, Enum):\n"
            "    def __new__(cls, name: str, side_effects: bool, pure: bool, control_flow: bool):\n"
            "        obj = str.__new__(cls, name)\n"
            "        obj._value_ = name\n"
            "        obj.side_effects = side_effects\n"
            "        obj.pure = pure\n"
            "        obj.control_flow = control_flow\n"
            "        return obj\n"
            "\n"
        )
        for function in data:
            name = function["name"]
            side_effects = function["sideEffects"]
            pure = function["pure"]
            control_flow = function["controlFlow"]
            out.write(f'    {name} = ("{name}", {side_effects}, {pure}, {control_flow})\n')


def block(name: str, f: TextIO, out: TextIO):
    data = json.load(f)
    out.write(f"class {name}Block(_Block, Enum):\n")
    for block in data:
        name = block["name"]
        id_ = block["id"]
        readable = block["readable"]
        writable = block["writable"]
        out.write(
            f"    {name} = ({id_}, {{{', '.join(f'"{e}"' for e in readable)}}}, {{{
                ', '.join(f'"{e}"' for e in writable)
            }}})\n"
        )


def blocks():
    with (out_dir / "blocks.py").open("w", encoding="utf-8") as out:
        out.write(
            "from enum import Enum\n"
            "\n"
            "\n"
            "class _Block(int):\n"
            "    def __new__(cls, id_: int, readable: set[str], writable: set[str]):\n"
            "        obj = int.__new__(cls, id_)\n"
            "        obj.readable = readable\n"
            "        obj.writable = writable\n"
            "        return obj\n"
        )
        for name, file in block_files.items():
            out.write("\n\n")
            with (
                (runtimes_dir / file).open("r", encoding="utf-8") as f,
            ):
                block(name, f, out)
        out.write(f"\n\ntype Block = {' | '.join(f'{name}Block' for name in block_files)}\n")


def main():
    out_dir.mkdir(exist_ok=True)
    functions()
    blocks()


if __name__ == "__main__":
    main()



from itertools import starmap

from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.containers import ArrayMap, Pair, VarArray
from sonolus.script.debug import assert_false, assert_true
from tests.script.conftest import run_and_validate

ints = st.integers(min_value=-999, max_value=999)
maps = st.dictionaries(ints, ints, min_size=1, max_size=20)


@st.composite
def map_and_key(draw):
    values = draw(maps)
    key = draw(st.sampled_from(list(values.keys())))
    return values, key


@st.composite
def map_and_missing_key(draw):
    values = draw(maps)
    key = draw(ints.filter(lambda x: x not in values))
    return values, key


@given(map_and_key())
def test_insertion(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return am[key]

    assert run_and_validate(fn) == values[key]


@given(map_and_key(), ints)
def test_update(args, new_value):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        am[key] = new_value
        return am[key]

    assert run_and_validate(fn) == new_value


@given(maps)
def test_keys(values):
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return am.keys()

    assert sorted(run_and_validate(fn)) == sorted(values.keys())


@given(maps)
def test_values(values):
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return am.values()

    assert sorted(run_and_validate(fn)) == sorted(values.values())


@given(maps)
def test_items(values):
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return am.items()

    assert sorted(run_and_validate(fn)) == sorted(values.items())


@given(map_and_key())
def test_pop(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)
    target_value = values[key]
    target_values = {k: v for k, v in values.items() if k != key}

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        assert_true(am.pop(key) == target_value)
        assert_false(key in am)
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(target_values.items())


@given(map_and_missing_key(), ints)
def test_insert_pop_round_trip(args, new_value):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count + 1].new()
        for pair in pairs:
            am[pair.first] = pair.second
        am[key] = new_value
        assert_true(am.pop(key) == new_value)
        assert_false(key in am)
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(values.items())


@given(map_and_key())
def test_pop_insert_round_trip(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)
    target_value = values[key]

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        assert_true(am.pop(key) == target_value)
        am[key] = target_value
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(values.items())


@given(map_and_key())
def test_delitem(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)
    target_values = {k: v for k, v in values.items() if k != key}

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        del am[key]
        assert_false(key in am)
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(target_values.items())


@given(map_and_missing_key(), ints)
def test_insert_delitem_round_trip(args, new_value):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count + 1].new()
        for pair in pairs:
            am[pair.first] = pair.second
        am[key] = new_value
        del am[key]
        assert_false(key in am)
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(values.items())


@given(map_and_key())
def test_delitem_insert_round_trip(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)
    target_value = values[key]

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        del am[key]
        am[key] = target_value
        return am

    assert sorted(run_and_validate(fn).items()) == sorted(values.items())


@given(maps)
def test_size(values):
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for i, pair in enumerate(pairs):
            assert_false(am.is_full())
            am[pair.first] = pair.second
            assert_true(len(am) == i + 1)
        keys = VarArray[int, count].new()
        for key in am:
            keys.append(key)
        assert_true(len(keys) == count)
        assert_true(am.is_full())
        for i, key in enumerate(keys):
            am.pop(key)
            assert_true(len(am) == count - i - 1)
            assert_false(am.is_full())
        return am

    assert len(run_and_validate(fn)) == 0


@given(maps)
def test_delitem_size(values):
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for i, pair in enumerate(pairs):
            assert_false(am.is_full())
            am[pair.first] = pair.second
            assert_true(len(am) == i + 1)
        keys = VarArray[int, count].new()
        for key in am:
            keys.append(key)
        assert_true(len(keys) == count)
        assert_true(am.is_full())
        for i, key in enumerate(keys):
            del am[key]
            assert_true(len(am) == count - i - 1)
            assert_false(am.is_full())
        return am

    assert len(run_and_validate(fn)) == 0


@given(map_and_key())
def test_contains_existing(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return key in am

    assert run_and_validate(fn)


@given(map_and_missing_key())
def test_contains_missing(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return key in am

    assert not run_and_validate(fn)


@given(map_and_key())
def test_not_contains_existing(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return key not in am

    assert not run_and_validate(fn)


@given(map_and_missing_key())
def test_not_contains_missing(args):
    values, key = args
    pairs = Array(*starmap(Pair, values.items()))
    count = len(values)

    def fn():
        am = ArrayMap[int, int, count].new()
        for pair in pairs:
            am[pair.first] = pair.second
        return key not in am

    assert run_and_validate(fn)


def test_array_map_truthiness_empty():
    def fn():
        x = ArrayMap[int, int, 5].new()
        return 1 if x else 0

    assert run_and_validate(fn) == 0


def test_array_map_truthiness_non_empty():
    def fn():
        x = ArrayMap[int, int, 5].new()
        x[1] = 100
        return 1 if x else 0

    assert run_and_validate(fn) == 1



from collections.abc import Iterable
from typing import overload

from sonolus.script.array import Array
from sonolus.script.array_like import ArrayLike
from sonolus.script.internal.context import ctx
from sonolus.script.internal.dict_impl import DictImpl
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.math_impls import MATH_BUILTIN_IMPLS, _trunc
from sonolus.script.internal.random import RANDOM_BUILTIN_IMPLS
from sonolus.script.internal.range import Range
from sonolus.script.internal.tuple_impl import TupleImpl
from sonolus.script.internal.value import Value
from sonolus.script.iterator import (
    SonolusIterator,
    _EmptyIterator,
    _Enumerator,
    _FilteringIterator,
    _MappingIterator,
    _Zipper,
)
from sonolus.script.num import Num, _is_num


@meta_fn
def _isinstance(value, type_):
    value = validate_value(value)
    type_ = validate_value(type_)._as_py_()
    if type_ is dict:
        return isinstance(value, DictImpl)
    if type_ is tuple:
        return isinstance(value, TupleImpl)
    if type_ in {_int, _float, _bool}:
        raise TypeError("Instance check against int, float, or bool is not supported, use Num instead")
    if not (isinstance(type_, type) and (issubclass(type_, Value) or getattr(type_, "_allow_instance_check_", False))):
        raise TypeError(f"Unsupported type: {type_} for isinstance")
    return validate_value(isinstance(value, type_))


@meta_fn
def _len(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__len__"):
        raise TypeError(f"object of type '{type(value).__name__}' has no len()")
    return compile_and_call(value.__len__)


@meta_fn
def _enumerate(iterable, start=0):
    from sonolus.backend.visitor import compile_and_call

    iterable = validate_value(iterable)
    if isinstance(iterable, TupleImpl):
        return TupleImpl._accept_(tuple((start + i, value) for i, value in enumerate(iterable._as_py_(), start=start)))
    elif not hasattr(iterable, "__iter__"):
        raise TypeError(f"'{type(iterable).__name__}' object is not iterable")
    elif isinstance(iterable, ArrayLike):
        return compile_and_call(iterable._enumerate_, start)
    else:
        iterator = compile_and_call(iterable.__iter__)
        if not isinstance(iterator, SonolusIterator):
            raise TypeError("Only subclasses of SonolusIterator are supported as iterators")
        return _Enumerator(0, start, iterator)


@meta_fn
def _reversed(iterable):
    from sonolus.backend.visitor import compile_and_call

    iterable = validate_value(iterable)
    if not isinstance(iterable, ArrayLike):
        raise TypeError(f"Unsupported type: {type(iterable)} for reversed")
    return compile_and_call(iterable.__reversed__)


@meta_fn
def _zip(*iterables):
    from sonolus.backend.visitor import compile_and_call
    from sonolus.script.containers import Pair

    if not iterables:
        return _EmptyIterator()

    iterables = [validate_value(iterable) for iterable in iterables]
    if any(isinstance(iterable, TupleImpl) for iterable in iterables):
        if not all(isinstance(iterable, TupleImpl) for iterable in iterables):
            raise TypeError("Cannot mix tuples with other types in zip")
        return TupleImpl._accept_(tuple(zip(*(iterable.value for iterable in iterables), strict=False)))
    iterators = [compile_and_call(iterable.__iter__) for iterable in iterables]
    if not all(isinstance(iterator, SonolusIterator) for iterator in iterators):
        raise TypeError("Only subclasses of SonolusIterator are supported as iterators")
    v = iterators.pop()
    while iterators:
        v = Pair(iterators.pop(), v)
    return _Zipper(v)


@meta_fn
def _abs(value):
    from sonolus.backend.visitor import compile_and_call

    value = validate_value(value)
    if not hasattr(value, "__abs__"):
        raise TypeError(f"bad operand type for abs(): '{type(value).__name__}'")
    return compile_and_call(value.__abs__)


def _identity(value):
    return value


@overload
def _max[T](iterable: Iterable[T], *, key: callable = ...) -> T: ...


@overload
def _max[T](a: T, b: T, *args: T, key: callable = ...) -> T: ...


@meta_fn
def _max(*args, key: callable = _identity):
    from sonolus.backend.visitor import compile_and_call

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to max")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return compile_and_call(iterable._max_, key=key)
        elif isinstance(iterable, TupleImpl) and all(_is_num(v) for v in iterable.value):
            return compile_and_call(Array(*iterable.value)._max_, key=key)
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for max")
    else:
        if not all(_is_num(arg) for arg in args):
            raise TypeError("Arguments to max must be numbers")
        if ctx():
            result = compile_and_call(_max2, args[0], args[1], key=key)
            for arg in args[2:]:
                result = compile_and_call(_max2, result, arg, key=key)
            return result
        else:
            return max(arg._as_py_() for arg in args)


def _max2(a, b, key=_identity):
    if key(a) > key(b):
        return a
    else:
        return b


@overload
def _min[T](iterable: Iterable[T], *, key: callable = ...) -> T: ...


@overload
def _min[T](a: T, b: T, *args: T, key: callable = ...) -> T: ...


@meta_fn
def _min(*args, key: callable = _identity):
    from sonolus.backend.visitor import compile_and_call

    args = tuple(validate_value(arg) for arg in args)
    if len(args) == 0:
        raise ValueError("Expected at least one argument to min")
    elif len(args) == 1:
        (iterable,) = args
        if isinstance(iterable, ArrayLike):
            return compile_and_call(iterable._min_, key=key)
        elif isinstance(iterable, TupleImpl) and all(_is_num(v) for v in iterable.value):
            return compile_and_call(Array(*iterable.value)._min_, key=key)
        else:
            raise TypeError(f"Unsupported type: {type(iterable)} for min")
    else:
        if not all(_is_num(arg) for arg in args):
            raise TypeError("Arguments to min must be numbers")
        if ctx():
            result = compile_and_call(_min2, args[0], args[1], key=key)
            for arg in args[2:]:
                result = compile_and_call(_min2, result, arg, key=key)
            return result
        else:
            return min(arg._as_py_() for arg in args)


def _min2(a, b, key=_identity):
    if key(a) < key(b):
        return a
    else:
        return b


@meta_fn
def _callable(value):
    return callable(value)


def _map(fn, iterable, *iterables):
    return _MappingIterator(lambda args: fn(*args), zip(iterable, *iterables))  # noqa: B905


def _filter(fn, iterable):
    if fn is None:
        fn = _identity
    return _FilteringIterator(fn, iterable.__iter__())  # noqa: PLC2801


@meta_fn
def _int(value=0):
    value = validate_value(value)
    if not _is_num(value):
        raise TypeError("Only numeric arguments to int() are supported")
    return _trunc(value)


@meta_fn
def _float(value=0.0):
    value = validate_value(value)
    if not _is_num(value):
        raise TypeError("Only numeric arguments to float() are supported")
    return value


def _bool(value=False):
    # Relies on the compiler to perform the conversion in a boolean context
    if value:  # noqa: SIM103
        return True
    else:
        return False


_int._type_mapping_ = Num
_float._type_mapping_ = Num
_bool._type_mapping_ = Num


def _any(iterable):
    for value in iterable:  # noqa: SIM110
        if value:
            return True
    return False


def _all(iterable):
    for value in iterable:  # noqa: SIM110
        if not value:
            return False
    return True


# classmethod, property, staticmethod are supported as decorators, but not within functions

BUILTIN_IMPLS = {
    id(abs): _abs,
    id(all): _all,
    id(any): _any,
    id(bool): _bool,
    id(callable): _callable,
    id(enumerate): _enumerate,
    id(filter): _filter,
    id(float): _float,
    id(int): _int,
    id(isinstance): _isinstance,
    id(len): _len,
    id(map): _map,
    id(max): _max,
    id(min): _min,
    id(range): Range,
    id(reversed): _reversed,
    id(zip): _zip,
    **MATH_BUILTIN_IMPLS,  # Includes round
    **RANDOM_BUILTIN_IMPLS,
}



"""Exception hook to filter out compiler internal frames from tracebacks."""

import sys
from types import TracebackType


def should_filter_traceback(tb: TracebackType | None) -> bool:
    return tb is not None and (
        tb.tb_frame.f_globals.get("_filter_traceback_", False) or should_filter_traceback(tb.tb_next)
    )


def is_compiler_internal(tb: TracebackType):
    return tb.tb_frame.f_locals.get("_compiler_internal_", False) or tb.tb_frame.f_globals.get(
        "_compiler_internal_", False
    )


def filter_traceback(tb: TracebackType | None) -> TracebackType | None:
    if tb is None:
        return None
    if is_compiler_internal(tb):
        return filter_traceback(tb.tb_next)
    tb.tb_next = filter_traceback(tb.tb_next)
    return tb


def excepthook(exc, value, tb):
    import traceback

    if should_filter_traceback(tb):
        tb = filter_traceback(tb)
    traceback.print_exception(exc, value, tb)


def install_excepthook():
    sys.excepthook = excepthook



# ruff: noqa: N802
import ast
import inspect
from collections.abc import Callable
from functools import cache
from pathlib import Path


@cache
def get_function(fn: Callable) -> tuple[str, ast.FunctionDef]:
    # This preserves both line number and column number in the returned node
    source_file = inspect.getsourcefile(fn)
    _, start_line = inspect.getsourcelines(fn)
    base_tree = get_tree_from_file(source_file)
    return source_file, find_function(base_tree, start_line)


@cache
def get_tree_from_file(file: str | Path) -> ast.Module:
    return ast.parse(Path(file).read_text(encoding="utf-8"))


class FindFunction(ast.NodeVisitor):
    def __init__(self, line):
        self.line = line
        self.node: ast.FunctionDef | None = None

    def visit_FunctionDef(self, node: ast.FunctionDef):
        if node.lineno == self.line or (
            node.decorator_list and (node.decorator_list[-1].end_lineno <= self.line <= node.lineno)
        ):
            self.node = node
        else:
            self.generic_visit(node)

    def visit_Lambda(self, node: ast.Lambda):
        if node.lineno == self.line:
            if self.node is not None:
                raise ValueError("Multiple functions defined on the same line are not supported")
            self.node = node
        else:
            self.generic_visit(node)


def find_function(tree: ast.Module, line: int):
    visitor = FindFunction(line)
    visitor.visit(tree)
    return visitor.node


class ScanWrites(ast.NodeVisitor):
    def __init__(self):
        self.writes = []

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store | ast.Delete):
            self.writes.append(node.id)


def scan_writes(node: ast.AST) -> set[str]:
    visitor = ScanWrites()
    visitor.visit(node)
    return set(visitor.writes)



# ruff: noqa: E501
import math

from sonolus.backend.ops import Op
from sonolus.script.internal.native import native_function
from sonolus.script.interval import clamp


@native_function(Op.EaseInBack)
def ease_in_back(x: float) -> float:
    """Interpolate between 0 and 1, starting slow and ending fast, overshooting below 0 at the start."""
    x = clamp(x, 0, 1)
    c1 = 1.70158
    c3 = c1 + 1
    return c3 * x**3 - c1 * x**2


@native_function(Op.EaseOutBack)
def ease_out_back(x: float) -> float:
    """Interpolate between 0 and 1, starting fast and ending slow, overshooting above 1 at the end."""
    x = clamp(x, 0, 1)
    c1 = 1.70158
    c3 = c1 + 1
    return 1 + c3 * (x - 1) ** 3 + c1 * (x - 1) ** 2


@native_function(Op.EaseInOutBack)
def ease_in_out_back(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending slow with overshooting, fast in the middle."""
    x = clamp(x, 0, 1)
    c1 = 1.70158
    c2 = c1 * 1.525
    if x < 0.5:
        return ((2 * x) ** 2 * ((c2 + 1) * 2 * x - c2)) / 2
    else:
        return ((2 * x - 2) ** 2 * ((c2 + 1) * (2 * x - 2) + c2) + 2) / 2


@native_function(Op.EaseOutInBack)
def ease_out_in_back(x: float) -> float:
    """Interpolate between 0 and 1, fast at the start and end, slow in the middle with overshooting."""
    x = clamp(x, 0, 1)
    c1 = 1.70158
    c3 = c1 + 1
    if x < 0.5:
        return (1 + c3 * (2 * x - 1) ** 3 + c1 * (2 * x - 1) ** 2) / 2
    else:
        return (c3 * (2 * x - 1) ** 3 - c1 * (2 * x - 1) ** 2) / 2 + 0.5


@native_function(Op.EaseInCirc)
def ease_in_circ(x: float) -> float:
    """Interpolate between 0 and 1, starting slow and ending very fast."""
    x = clamp(x, 0, 1)
    return 1 - math.sqrt(1 - x**2)


@native_function(Op.EaseOutCirc)
def ease_out_circ(x: float) -> float:
    """Interpolate between 0 and 1, starting very fast and ending slow."""
    x = clamp(x, 0, 1)
    return math.sqrt(1 - (x - 1) ** 2)


@native_function(Op.EaseInOutCirc)
def ease_in_out_circ(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending slow, very fast in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return (1 - math.sqrt(1 - (2 * x) ** 2)) / 2
    else:
        return (math.sqrt(1 - (2 * x - 2) ** 2) + 1) / 2


@native_function(Op.EaseOutInCirc)
def ease_out_in_circ(x: float) -> float:
    """Interpolate between 0 and 1, very fast at the start and end, slow in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return math.sqrt(1 - (2 * x - 1) ** 2) / 2
    else:
        return (1 - math.sqrt(1 - (2 * x - 1) ** 2)) / 2 + 0.5


@native_function(Op.EaseInCubic)
def ease_in_cubic(x: float) -> float:
    """Interpolate between 0 and 1, starting slow and ending fast with cubic easing."""
    x = clamp(x, 0, 1)
    return x**3


@native_function(Op.EaseOutCubic)
def ease_out_cubic(x: float) -> float:
    """Interpolate between 0 and 1, starting fast and ending slow with cubic easing."""
    x = clamp(x, 0, 1)
    return 1 - (1 - x) ** 3


@native_function(Op.EaseInOutCubic)
def ease_in_out_cubic(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending slow with cubic easing, fast in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return 4 * x**3
    else:
        return 1 - (-2 * x + 2) ** 3 / 2


@native_function(Op.EaseOutInCubic)
def ease_out_in_cubic(x: float) -> float:
    """Interpolate between 0 and 1, fast at the start and end, slow in the middle with cubic easing."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return (1 - (1 - 2 * x) ** 3) / 2
    else:
        return ((2 * x - 1) ** 3) / 2 + 0.5


@native_function(Op.EaseInElastic)
def ease_in_elastic(x: float) -> float:
    """Interpolate between 0 and 1 with oscillations, starting slow and ending fast."""
    x = clamp(x, 0, 1)
    c4 = (2 * math.pi) / 3
    if x in {0, 1}:
        return x
    else:
        return -(2 ** (10 * x - 10)) * math.sin((x * 10 - 10.75) * c4)


@native_function(Op.EaseOutElastic)
def ease_out_elastic(x: float) -> float:
    """Interpolate between 0 and 1 with oscillations, starting fast and ending slow."""
    x = clamp(x, 0, 1)
    c4 = (2 * math.pi) / 3
    if x in {0, 1}:
        return x
    else:
        return 2 ** (-10 * x) * math.sin((x * 10 - 0.75) * c4) + 1


@native_function(Op.EaseInOutElastic)
def ease_in_out_elastic(x: float) -> float:
    """Interpolate between 0 and 1 with oscillations, slow at the start and end, fast in the middle."""
    x = clamp(x, 0, 1)
    c5 = (2 * math.pi) / 4.5
    if x in {0, 1}:
        return x
    elif x < 0.5:
        return -(2 ** (20 * x - 10) * math.sin((20 * x - 11.125) * c5)) / 2
    else:
        return (2 ** (-20 * x + 10) * math.sin((20 * x - 11.125) * c5)) / 2 + 1


@native_function(Op.EaseOutInElastic)
def ease_out_in_elastic(x: float) -> float:
    """Interpolate between 0 and 1 with oscillations, fast at the start and end, slow in the middle."""
    x = clamp(x, 0, 1)
    c4 = (2 * math.pi) / 3
    if x < 0.5:
        if x == 0:
            return 0
        else:
            return (2 ** (-20 * x + 10) * math.sin((20 * x - 0.75) * c4)) / 2 + 0.5
    elif x == 1:
        return 1
    else:
        return (-(2 ** (10 * (2 * x - 1) - 10)) * math.sin((20 * x - 10.75) * c4)) / 2 + 0.5


@native_function(Op.EaseInExpo)
def ease_in_expo(x: float) -> float:
    """Interpolate between 0 and 1, starting extremely slow and ending extremely fast."""
    x = clamp(x, 0, 1)
    return 0 if x == 0 else 2 ** (10 * x - 10)


@native_function(Op.EaseOutExpo)
def ease_out_expo(x: float) -> float:
    """Interpolate between 0 and 1, starting extremely fast and ending extremely slow."""
    x = clamp(x, 0, 1)
    return 1 if x == 1 else 1 - 2 ** (-10 * x)


@native_function(Op.EaseInOutExpo)
def ease_in_out_expo(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending extremely slow, fast in the middle."""
    x = clamp(x, 0, 1)
    if x in {0, 1}:
        return x
    elif x < 0.5:
        return 2 ** (20 * x - 10) / 2
    else:
        return (2 - 2 ** (-20 * x + 10)) / 2


@native_function(Op.EaseOutInExpo)
def ease_out_in_expo(x: float) -> float:
    """Interpolate between 0 and 1, extremely fast at the start and end, extremely slow in the middle."""
    x = clamp(x, 0, 1)
    if x in {0, 1}:
        return x
    elif x < 0.5:
        return (1 - 2 ** (-20 * x)) / 2
    else:
        return (2 ** (20 * x - 20)) / 2 + 0.5


@native_function(Op.EaseInQuad)
def ease_in_quad(x: float) -> float:
    """Interpolate between 0 and 1, starting slow and ending fast with quadratic easing."""
    x = clamp(x, 0, 1)
    return x**2


@native_function(Op.EaseOutQuad)
def ease_out_quad(x: float) -> float:
    """Interpolate between 0 and 1, starting fast and ending slow with quadratic easing."""
    x = clamp(x, 0, 1)
    return 1 - (1 - x) ** 2


@native_function(Op.EaseInOutQuad)
def ease_in_out_quad(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending slow with quadratic easing, fast in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return 2 * x**2
    else:
        return 1 - (-2 * x + 2) ** 2 / 2


@native_function(Op.EaseOutInQuad)
def ease_out_in_quad(x: float) -> float:
    """Interpolate between 0 and 1, fast at the start and end, slow in the middle with quadratic easing."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return (1 - (1 - 2 * x) ** 2) / 2
    else:
        return ((2 * x - 1) ** 2) / 2 + 0.5


@native_function(Op.EaseInQuart)
def ease_in_quart(x: float) -> float:
    """Interpolate between 0 and 1, starting very slow and ending very fast with quartic easing."""
    x = clamp(x, 0, 1)
    return x**4


@native_function(Op.EaseOutQuart)
def ease_out_quart(x: float) -> float:
    """Interpolate between 0 and 1, starting very fast and ending very slow with quartic easing."""
    x = clamp(x, 0, 1)
    return 1 - (1 - x) ** 4


@native_function(Op.EaseInOutQuart)
def ease_in_out_quart(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending very slow with quartic easing, very fast in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return 8 * x**4
    else:
        return 1 - (-2 * x + 2) ** 4 / 2


@native_function(Op.EaseOutInQuart)
def ease_out_in_quart(x: float) -> float:
    """Interpolate between 0 and 1, very fast at the start and end, very slow in the middle with quartic easing."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return (1 - (1 - 2 * x) ** 4) / 2
    else:
        return ((2 * x - 1) ** 4) / 2 + 0.5


@native_function(Op.EaseInQuint)
def ease_in_quint(x: float) -> float:
    """Interpolate between 0 and 1, starting extremely slow and ending extremely fast with quintic easing."""
    x = clamp(x, 0, 1)
    return x**5


@native_function(Op.EaseOutQuint)
def ease_out_quint(x: float) -> float:
    """Interpolate between 0 and 1, starting extremely fast and ending extremely slow with quintic easing."""
    x = clamp(x, 0, 1)
    return 1 - (1 - x) ** 5


@native_function(Op.EaseInOutQuint)
def ease_in_out_quint(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending extremely slow with quintic easing, extremely fast in the middle."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return 16 * x**5
    else:
        return 1 - (-2 * x + 2) ** 5 / 2


@native_function(Op.EaseOutInQuint)
def ease_out_in_quint(x: float) -> float:
    """Interpolate between 0 and 1, extremely fast at the start and end, extremely slow in the middle with quintic easing."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return (1 - (1 - 2 * x) ** 5) / 2
    else:
        return ((2 * x - 1) ** 5) / 2 + 0.5


@native_function(Op.EaseInSine)
def ease_in_sine(x: float) -> float:
    """Interpolate between 0 and 1, starting slow and ending fast with sine easing."""
    x = clamp(x, 0, 1)
    return 1 - math.cos((x * math.pi) / 2)


@native_function(Op.EaseOutSine)
def ease_out_sine(x: float) -> float:
    """Interpolate between 0 and 1, starting fast and ending slow with sine easing."""
    x = clamp(x, 0, 1)
    return math.sin((x * math.pi) / 2)


@native_function(Op.EaseInOutSine)
def ease_in_out_sine(x: float) -> float:
    """Interpolate between 0 and 1, starting and ending slow with sine easing, fast in the middle."""
    x = clamp(x, 0, 1)
    return -(math.cos(math.pi * x) - 1) / 2


@native_function(Op.EaseOutInSine)
def ease_out_in_sine(x: float) -> float:
    """Interpolate between 0 and 1, fast at the start and end, slow in the middle with sine easing."""
    x = clamp(x, 0, 1)
    if x < 0.5:
        return math.sin(math.pi * x) / 2
    else:
        return (1 - math.cos(math.pi * x)) / 2


def linstep(x: float) -> float:
    """Linear interpolation between 0 and 1."""
    return clamp(x, 0.0, 1.0)


def smoothstep(x: float) -> float:
    """Interpolate between 0 and 1 using smoothstep."""
    x = clamp(x, 0.0, 1.0)
    return x * x * (3 - 2 * x)


def smootherstep(x: float) -> float:
    """Interpolate between 0 and 1 using smootherstep."""
    x = clamp(x, 0.0, 1.0)
    return x * x * x * (x * (x * 6 - 15) + 10)


def step_start(x: float) -> float:
    """Step function returning 1.0 if x > 0, otherwise 0.0."""
    return 1.0 if x > 0 else 0.0


def step_end(x: float) -> float:
    """Step function returning 1.0 if x >= 1, otherwise 0.0."""
    return 1.0 if x >= 1 else 0.0



from __future__ import annotations

from abc import abstractmethod
from typing import Any

from sonolus.script.internal.impl import meta_fn
from sonolus.script.record import Record


class SonolusIterator[T]:
    """Base class for Sonolus iterators.

    This class is used to define custom iterators that can be used in Sonolus.py.

    Inheritors must implement the `has_next`, `get`, and `advance` methods.
    The `__next__` and `__iter__` methods are implemented by default.

    Usage:
        ```python
        class MyIterator(Record, SonolusIterator):
            def has_next(self) -> bool:
                ...

            def get(self) -> Any:
                ...

            def advance(self):
                ...
        ```
    """

    _allow_instance_check_ = True

    def next(self) -> T:
        result = self.get()
        self.advance()
        return result

    @abstractmethod
    def has_next(self) -> bool:
        """Return whether the iterator has more elements."""
        raise NotImplementedError

    @abstractmethod
    def get(self) -> T:
        """Return the next element of the iterator.

        May be called multiple times before calling `advance`.

        Must not be called if `has_next` returns `False`.
        """
        raise NotImplementedError

    @abstractmethod
    def advance(self):
        """Advance the iterator to the next element.

        Must not be called if `has_next` returns `False`.
        """
        raise NotImplementedError

    def __next__(self) -> T:
        if not self.has_next():
            raise StopIteration
        return self.next()

    def __iter__(self) -> SonolusIterator[T]:
        return self


class _Enumerator[V: SonolusIterator](Record, SonolusIterator):
    i: int
    offset: int
    iterator: V

    def has_next(self) -> bool:
        return self.iterator.has_next()

    def get(self) -> tuple[int, Any]:
        return self.i + self.offset, self.iterator.get()

    def advance(self):
        self.i += 1
        self.iterator.advance()


class _Zipper[T](Record, SonolusIterator):
    # Can be a, Pair[a, b], Pair[a, Pair[b, c]], etc.
    iterators: T

    @meta_fn
    def has_next(self) -> bool:
        from sonolus.backend.visitor import compile_and_call

        return compile_and_call(self._has_next, self._get_iterators())

    def _get_iterators(self) -> tuple[SonolusIterator, ...]:
        from sonolus.script.containers import Pair

        iterators = []
        v = self.iterators
        while isinstance(v, Pair):
            iterators.append(v.first)
            v = v.second
        iterators.append(v)
        return tuple(iterators)

    def _has_next(self, iterators: tuple[SonolusIterator, ...]) -> bool:
        for iterator in iterators:  # noqa: SIM110
            if not iterator.has_next():
                return False
        return True

    @meta_fn
    def get(self) -> tuple[Any, ...]:
        from sonolus.backend.visitor import compile_and_call

        return tuple(compile_and_call(iterator.get) for iterator in self._get_iterators())

    @meta_fn
    def advance(self):
        from sonolus.backend.visitor import compile_and_call

        for iterator in self._get_iterators():
            compile_and_call(iterator.advance)


class _EmptyIterator(Record, SonolusIterator):
    def has_next(self) -> bool:
        return False

    def get(self) -> Any:
        return None

    def advance(self):
        pass


class _MappingIterator[T, Fn](Record, SonolusIterator):
    fn: Fn
    iterator: T

    def has_next(self) -> bool:
        return self.iterator.has_next()

    def get(self) -> Any:
        return self.fn(self.iterator.get())

    def advance(self):
        self.iterator.advance()


class _FilteringIterator[T, Fn](Record, SonolusIterator):
    fn: Fn
    iterator: T

    def has_next(self) -> bool:
        while self.iterator.has_next():
            if self.fn(self.iterator.get()):
                return True
            self.iterator.advance()
        return False

    def get(self) -> Any:
        return self.iterator.get()

    def advance(self):
        self.iterator.advance()


class _ChainingIterator[T](Record, SonolusIterator):
    iterator: T

    def has_next(self) -> bool:
        return self.iterator.has_next()

    def get(self) -> Any:
        return self.iterator.get().get()

    def advance(self):
        self.iterator.get().advance()
        while not self.iterator.get().has_next():
            self.iterator.advance()
            if not self.iterator.has_next():
                break



# ruff: noqa
"""Test cases intended to cover more complex control flow."""

import random

import pytest
from hypothesis import given
from hypothesis import strategies as st

from sonolus.script.array import Array
from sonolus.script.debug import debug_log
from sonolus.script.internal.error import CompilationError
from tests.script.conftest import run_compiled
from tests.script.conftest import run_and_validate
from tests.script.test_record import Pair

ints = st.integers(min_value=-999999, max_value=999999)
small_ints = st.integers(min_value=-2, max_value=2)
floats = st.floats(min_value=-999999, max_value=999999, allow_infinity=False, allow_nan=False)
divisor_floats = floats.filter(lambda x: abs(x) > 1e-6)


def test_loop_with_side_effects():
    def fn():
        for i in range(10):
            # This can't be inlined at all since it would change the behavior
            x = debug_log(i)
            if i % 2 == 0:
                debug_log(x)
                # We can inline this, but we need to remove the original statement
                y = debug_log(i)
                debug_log(y)

    run_and_validate(fn)


def test_nested_loops_with_breaks():
    def fn():
        for i in range(5):
            debug_log(i)
            for j in range(5):
                x = debug_log(j)
                if x > 2:
                    break
                debug_log(x * i)
            debug_log(-i)

    run_and_validate(fn)


def test_conditional_assignments():
    def fn():
        x = 0
        for i in range(5):
            # Test conditional assignments with side effects
            x = debug_log(i) if i % 2 == 0 else debug_log(-i)
            debug_log(x)

            # Test multiple branches with side effects
            if i > 2:
                y = debug_log(i * 2)
            elif i > 1:
                y = debug_log(i * 3)
            else:
                y = debug_log(i * 4)
            debug_log(y)

    run_and_validate(fn)


def test_loop_with_continue():
    def fn():
        for i in range(10):
            x = debug_log(i)
            if i % 3 == 0:
                continue
            debug_log(x)
            if i % 2 == 0:
                y = debug_log(i * 2)
                debug_log(y)

    run_and_validate(fn)


def test_nested_conditionals():
    def fn():
        for i in range(5):
            x = debug_log(i)
            if i > 0:
                if i > 2:
                    y = debug_log(x * 2)
                    if i > 3:
                        debug_log(y * 2)
                    debug_log(y)
                debug_log(x)
            debug_log(-i)

    run_and_validate(fn)


def test_variable_reassignment():
    def fn():
        x = debug_log(0)
        for i in range(5):
            debug_log(x)
            x = debug_log(i)  # Reassign x with side effect
            if i % 2 == 0:
                x = debug_log(x * 2)  # Reassign again
            debug_log(x)

    run_and_validate(fn)


def test_early_returns():
    def fn():
        for i in range(10):
            x = debug_log(i)
            if i > 5:
                debug_log(-1)
                return
            debug_log(x)
            if i % 2 == 0:
                y = debug_log(i * 2)
                debug_log(y)

    run_and_validate(fn)


def test_loop_variable_dependencies():
    def fn():
        prev = debug_log(0)
        curr = debug_log(1)
        for i in range(5):
            debug_log(prev)
            debug_log(curr)
            temp = debug_log(curr)
            curr = debug_log(prev + curr)
            prev = debug_log(temp)
            debug_log(i)

    run_and_validate(fn)


def test_pair_mutations_in_loop():
    def fn():
        p = Pair(0, 0)
        for i in range(5):
            # Test mutation of first field with side effects
            p.first = debug_log(i)
            debug_log(p.first)

            # Test mutation of second field in conditional
            if i % 2 == 0:
                p.second = debug_log(i * 2)
                debug_log(p.second)

            # Test reading after mutation
            debug_log(p.first + p.second)
        return p

    run_and_validate(fn)


def test_pair_copy_from_operator():
    def fn():
        p = Pair(0, 0)
        for i in range(5):
            debug_log(p.first)
            # Test copy-from with side effects in constructor
            p @= Pair(debug_log(i), debug_log(i * 2))
            debug_log(p.second)

            if i % 2 == 0:
                # Test nested copy-from operations
                temp = Pair(debug_log(i * 3), debug_log(i * 4))
                p @= temp
                debug_log(p.first)
        return p

    run_and_validate(fn)


def test_pair_conditional_mutations():
    def fn():
        p = Pair(0, 0)
        for i in range(5):
            # Test conditional mutations with side effects
            p.first = debug_log(i) if i % 2 == 0 else debug_log(-i)
            debug_log(p.first)

            # Test multiple mutation branches
            if i > 2:
                p.second = debug_log(i * 2)
            elif i > 1:
                p.second = debug_log(i * 3)
            else:
                p @= Pair(debug_log(i), debug_log(-i))
            debug_log(p.second)
        return p

    run_and_validate(fn)


def test_pair_nested_mutations():
    def fn():
        p1 = Pair(0, 0)
        p2 = Pair(1, 1)
        for i in range(5):
            debug_log(p1.first)
            if i % 2 == 0:
                # Test interleaved mutations between two pairs
                p1.first = debug_log(i)
                p2.second = debug_log(i * 2)
                p1 @= p2
                debug_log(p1.second)
            else:
                # Test copy followed by mutation
                p2 @= p1
                p2.first = debug_log(-i)
                debug_log(p2.first)
            debug_log(p1.first + p2.second)
        return p1

    run_and_validate(fn)


def test_pair_early_return_with_mutations():
    def fn():
        p = Pair(0, 0)
        for i in range(10):
            p.first = debug_log(i)
            if i > 5:
                p.second = debug_log(-1)
                debug_log(p.second)
                return p
            debug_log(p.first)
            if i % 2 == 0:
                p @= Pair(debug_log(i * 2), debug_log(i * 3))
                debug_log(p.first)
        return p

    run_and_validate(fn)


def test_random_multi_use():
    def add(a, b):
        return a + b

    # Random has no side effects, but is impure, so we need to test that optimizations don't break it.
    def fn():
        a = random.uniform(1, 10)
        b = add(a, Pair(a, a).first)
        c = add(b, -2 * a)
        return c == 0

    for _ in range(100):
        run_and_validate(fn)


def test_switch_with_integer_cases():
    def fn():
        for i in range(5):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 1:
                    debug_log(11)
                case 2:
                    debug_log(22)
                case 3:
                    debug_log(33)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_default():
    def fn():
        for i in range(5):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 1:
                    debug_log(11)
                case 2:
                    debug_log(22)
                case 3:
                    debug_log(33)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_offset_integer_cases():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 3:
                    debug_log(22)
                case 4:
                    debug_log(33)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_offset_integer_cases_and_default():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 3:
                    debug_log(22)
                case 4:
                    debug_log(33)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_stride():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 4:
                    debug_log(22)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_stride_and_default():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 4:
                    debug_log(22)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_stride_and_offset():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 3:
                    debug_log(11)
                case 5:
                    debug_log(22)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_stride_and_offset_and_default():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 3:
                    debug_log(11)
                case 5:
                    debug_log(22)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_variable_stride():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 5:
                    debug_log(22)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_variable_stride_and_default():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 0:
                    debug_log(0)
                case 2:
                    debug_log(11)
                case 5:
                    debug_log(22)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_variable_stride_and_offset():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 3:
                    debug_log(11)
                case 6:
                    debug_log(22)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_integer_cases_and_variable_stride_and_offset_and_default():
    def fn():
        for i in range(10):
            debug_log(i)
            match i:
                case 1:
                    debug_log(0)
                case 3:
                    debug_log(11)
                case 6:
                    debug_log(22)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_float_cases():
    def fn():
        for i in range(10):
            debug_log(i)
            match i / 2:
                case 0.0:
                    debug_log(0)
                case 0.5:
                    debug_log(11)
                case 1.0:
                    debug_log(22)
                case 3.0:
                    debug_log(33)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_float_cases_and_default():
    def fn():
        for i in range(5):
            debug_log(i)
            match i / 2:
                case 0.0:
                    debug_log(0)
                case 0.5:
                    debug_log(11)
                case 1.0:
                    debug_log(22)
                case 1.5:
                    debug_log(33)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_out_of_order_integer_cases():
    def fn():
        for i in range(5):
            debug_log(i)
            match i:
                case 2:
                    debug_log(0)
                case 0:
                    debug_log(11)
                case 3:
                    debug_log(22)
                case 1:
                    debug_log(33)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_out_of_order_integer_cases_and_default():
    def fn():
        for i in range(5):
            debug_log(i)
            match i:
                case 2:
                    debug_log(0)
                case 0:
                    debug_log(11)
                case 3:
                    debug_log(22)
                case 1:
                    debug_log(33)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_out_of_order_float_cases():
    def fn():
        for i in range(5):
            debug_log(i)
            match i / 2:
                case 1.0:
                    debug_log(0)
                case 0.0:
                    debug_log(11)
                case 1.5:
                    debug_log(22)
                case 0.5:
                    debug_log(33)
            debug_log(i)

    run_and_validate(fn)


def test_switch_with_out_of_order_float_cases_and_default():
    def fn():
        for i in range(5):
            debug_log(i)
            match i / 2:
                case 1.0:
                    debug_log(0)
                case 0.0:
                    debug_log(11)
                case 1.5:
                    debug_log(22)
                case 0.5:
                    debug_log(33)
                case _:
                    debug_log(-1)
            debug_log(i)

    run_and_validate(fn)


def test_while_else_taken():
    def fn():
        i = 0
        while i < 5:
            debug_log(i)
            i += 1
        debug_log(-1)

    run_and_validate(fn)


def test_while_else_not_taken():
    def fn():
        i = 0
        while i < 5:
            debug_log(i)
            i += 1
            if i == 3:
                break
        else:
            debug_log(-1)

    run_and_validate(fn)


def test_for_else_taken():
    def fn():
        for i in range(5):
            debug_log(i)
        debug_log(-1)

    run_and_validate(fn)


def test_for_else_not_taken():
    def fn():
        for i in range(5):
            debug_log(i)
            if i == 3:
                break
        else:
            debug_log(-1)

    run_and_validate(fn)


def black_box():
    # This really always returns True, but the optimizer doesn't know that,
    # so we can use it as a black box to prevent branches from being optimized away.
    return random.randrange(0, 1) == 0


def black_box_value(v: float | int) -> float | int:
    if black_box():
        return v
    return 0


def black_box_log(v: float | int) -> float | int:
    debug_log(v)
    return black_box_value(v)


def test_error_if_conflicting_definitions():
    def fn():
        x = Pair(1, 2)
        if black_box():
            x = Pair(3, 4)
        debug_log(x.first)

    with pytest.raises(CompilationError, match="conflicting definitions"):
        run_compiled(fn)


def test_error_while_conflicting_definitions():
    def fn():
        x = Pair(1, 2)
        while black_box():
            debug_log(x.first)
            x = Pair(3, 4)
        return 1

    with pytest.raises(CompilationError, match="conflicting definitions"):
        run_compiled(fn)


def test_error_for_conflicting_definitions():
    def fn():
        x = Pair(1, 2)
        for _ in range(5):
            debug_log(x.first)
            x = Pair(3, 4)
        return 1

    with pytest.raises(CompilationError, match="conflicting definitions"):
        run_compiled(fn)


def test_walrus_operator():
    def fn():
        x: int = 0
        while (y := x) < 5:
            debug_log(y)
            x += 1

    run_and_validate(fn)


def test_match_singletons():
    def m(x):
        match x:
            case None:
                return 0
            case _:
                return 1

    def fn():
        return Array(m(None), m(0))

    assert run_and_validate(fn) == Array(0, 1)


def test_match_true_not_supported():
    def m(x):
        match x:
            case True:
                return 0
            case _:
                return 1

    def fn():
        m(True)
        return 1

    with pytest.raises(CompilationError, match="not supported"):
        run_compiled(fn)


def test_match_false_not_supported():
    def m(x):
        match x:
            case False:
                return 0
            case _:
                return 1

    def fn():
        m(False)
        return 1

    with pytest.raises(CompilationError, match="not supported"):
        run_compiled(fn)


def test_match_int_not_supported():
    def m(x):
        match x:
            case int():
                return 0

    def fn():
        m(1)
        return 1

    with pytest.raises(CompilationError, match="not supported"):
        run_compiled(fn)


@given(small_ints, small_ints, small_ints)
def test_and(x, y, z):
    def fn():
        a = x and y and z
        b = black_box_log(x) and y and z
        c = x and black_box_log(y) and z
        d = x and y and black_box_log(z)
        e = black_box_log(x) and black_box_log(y) and z
        f = black_box_log(x) and y and black_box_log(z)
        g = x and black_box_log(y) and black_box_log(z)
        h = black_box_log(x) and black_box_log(y) and black_box_log(z)
        return Array(a, b, c, d, e, f, g, h)

    assert run_and_validate(fn) == Array(*(x and y and z for _ in range(8)))


@given(small_ints, small_ints, small_ints)
def test_or(x, y, z):
    def fn():
        a = x or y or z
        b = black_box_log(x) or y or z
        c = x or black_box_log(y) or z
        d = x or y or black_box_log(z)
        e = black_box_log(x) or black_box_log(y) or z
        f = black_box_log(x) or y or black_box_log(z)
        g = x or black_box_log(y) or black_box_log(z)
        h = black_box_log(x) or black_box_log(y) or black_box_log(z)
        return Array(a, b, c, d, e, f, g, h)

    assert run_and_validate(fn) == Array(*(x or y or z for _ in range(8)))


@given(small_ints, small_ints, small_ints)
def test_and_or(x, y, z):
    def fn():
        a = x and y or z
        b = black_box_log(x) and y or z
        c = x and black_box_log(y) or z
        d = x and y or black_box_log(z)
        e = black_box_log(x) and black_box_log(y) or z
        f = black_box_log(x) and y or black_box_log(z)
        g = x and black_box_log(y) or black_box_log(z)
        h = black_box_log(x) and black_box_log(y) or black_box_log(z)
        return Array(a, b, c, d, e, f, g, h)

    assert run_and_validate(fn) == Array(*(x and y or z for _ in range(8)))


@given(small_ints, small_ints, small_ints)
def test_or_and(x, y, z):
    def fn():
        a = x or y and z
        b = black_box_log(x) or y and z
        c = x or black_box_log(y) and z
        d = x or y and black_box_log(z)
        e = black_box_log(x) or black_box_log(y) and z
        f = black_box_log(x) or y and black_box_log(z)
        g = x or black_box_log(y) and black_box_log(z)
        h = black_box_log(x) or black_box_log(y) and black_box_log(z)
        return Array(a, b, c, d, e, f, g, h)

    assert run_and_validate(fn) == Array(*(x or y and z for _ in range(8)))


@given(small_ints, small_ints, small_ints)
def test_chained_comparison(x, y, z):
    def fn():
        a = x < y < z
        b = black_box_log(x) < y < z
        c = x < black_box_log(y) < z
        d = x < y < black_box_log(z)
        e = black_box_log(x) < black_box_log(y) < z
        f = black_box_log(x) < y < black_box_log(z)
        g = x < black_box_log(y) < black_box_log(z)
        h = black_box_log(x) < black_box_log(y) < black_box_log(z)
        return Array(a, b, c, d, e, f, g, h)

    assert run_and_validate(fn) == Array(*(x < y < z for _ in range(8)))


def test_while_true():
    def fn():
        debug_log(1)
        while True:
            debug_log(2)
            break
        else:
            debug_log(3)
        debug_log(4)

    run_and_validate(fn)


def test_while_false():
    def fn():
        debug_log(1)
        while False:
            debug_log(2)
        else:
            debug_log(3)
        debug_log(4)

    run_and_validate(fn)


def test_for_empty():
    def fn():
        debug_log(1)
        for _ in zip():
            debug_log(2)
        else:
            debug_log(3)
        debug_log(4)

    run_and_validate(fn)


def test_loop_with_aug_assign():
    def fn():
        a = 0
        b = 0
        while a < 5:
            debug_log(a + b)
            b = a
            a += 1

    for _ in range(100):
        run_and_validate(fn)


def test_break_in_nested_for_else():
    def fn():
        for _ in range(2):
            for _ in range(2):
                debug_log(1)
            else:
                debug_log(2)
                break
            debug_log(3)
        else:
            debug_log(4)
        debug_log(5)

    run_and_validate(fn)



import inspect
from typing import Annotated

_missing = object()


def get_field_specifiers(cls, *, skip: set[str] = frozenset(), globals=None, locals=None, eval_str=True):  # noqa: A002
    """Like inspect.get_annotations, but also turns class attributes into Annotated."""
    results = inspect.get_annotations(cls, globals=globals, locals=locals, eval_str=eval_str)
    for key, value in results.items():
        class_value = getattr(cls, key, _missing)
        if class_value is not _missing and key not in skip:
            results[key] = Annotated[value, class_value]
    for key, value in cls.__dict__.items():
        if (
            key not in results
            and key not in skip
            and not key.startswith("__")
            and not callable(value)
            and not hasattr(value, "__func__")
            and not isinstance(value, property)
        ):
            raise ValueError(f"Missing annotation for {cls.__name__}.{key}")
    return results



from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Self, final

from sonolus.backend.ir import IRConst, IRSet
from sonolus.backend.place import BlockPlace
from sonolus.script.array_like import ArrayLike, get_positive_index
from sonolus.script.debug import assert_unreachable
from sonolus.script.internal.context import ctx
from sonolus.script.internal.error import InternalError
from sonolus.script.internal.generic import GenericValue
from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.value import BackingSource, DataValue, Value
from sonolus.script.num import Num


class ArrayMeta(type):
    @meta_fn
    def __pos__[T](cls: type[T]) -> T:
        """Create a zero-initialized array instance."""
        return cls._zero_()


@final
class Array[T, Size](GenericValue, ArrayLike[T], metaclass=ArrayMeta):
    """A fixed size array of values.

    Usage:
        ```python
        array_1 = Array(1, 2, 3)
        array_2 = Array[int, 0]()
        array_3 = +Array[int, 3]  # Create a zero-initialized array
        ```
    """

    _value: list[T] | BlockPlace | BackingSource

    @classmethod
    def element_type(cls) -> type[T] | type[Value]:
        """Return the type of elements in this array type."""
        return cls.type_var_value(T)

    @classmethod
    def size(cls) -> int:
        """Return the size of this array type.

        On instances, use `len(array)` instead.
        """
        return cls.type_var_value(Size)

    def __new__(cls, *args: T) -> Array[T, Any]:
        if cls._type_args_ is None:
            values = [validate_value(arg) for arg in args]
            types = {type(value) for value in values}
            if len(types) == 0:
                raise ValueError(
                    f"{cls.__name__} constructor should be used with at least one value if type is not specified"
                )
            if len(types) > 1:
                raise TypeError(f"{cls.__name__} constructor should be used with values of the same type, got {types}")
            parameterized_cls = cls[types.pop(), len(args)]
        else:
            values = [cls.element_type()._accept_(arg) for arg in args]
            if len(args) != cls.size():
                raise ValueError(f"{cls.__name__} constructor should be used with {cls.size()} values, got {len(args)}")
            parameterized_cls = cls
        if ctx():
            place = ctx().alloc(size=parameterized_cls._size_())
            result: parameterized_cls = parameterized_cls._from_place_(place)
            result._copy_from_(parameterized_cls._with_value(values))
            return result
        else:
            return parameterized_cls._with_value([value._copy_() for value in values])

    def __init__(self, *args: T):
        super().__init__()

    @classmethod
    def _with_value(cls, value) -> Self:
        result = object.__new__(cls)
        result._value = value
        return result

    @classmethod
    def _size_(cls) -> int:
        return cls.size() * cls.element_type()._size_()

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Self:
        return cls._with_value(source)

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        return cls._with_value(place)

    @classmethod
    def _accepts_(cls, value: Any) -> bool:
        return isinstance(value, cls)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept value {value} as {cls.__name__}")
        return value

    def _is_py_(self) -> bool:
        return isinstance(self._value, list)

    def _as_py_(self) -> Any:
        if not self._is_py_():
            raise ValueError("Not a python value")
        return self

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        iterator = iter(values)
        return cls(*(cls.element_type()._from_list_(iterator) for _ in range(cls.size())))

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        match self._value:
            case list():
                return [entry for value in self._value for entry in value._to_list_(level_refs)]
            case BlockPlace():
                return [
                    entry
                    for i in range(self.size())
                    for entry in self.element_type()
                    ._from_place_(self._value.add_offset(i * self.element_type()._size_()))
                    ._to_list_()
                ]
            case backing_source if callable(backing_source):
                return [backing_source(IRConst(i)) for i in range(self.size() * self.element_type()._size_())]
            case _:
                assert_unreachable()

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return [entry for i in range(cls.size()) for entry in cls.element_type()._flat_keys_(f"{prefix}[{i}]")]

    def _get_(self) -> Self:
        return self

    def _set_(self, value: Self):
        raise TypeError("Array does not support _set_")

    def _copy_from_(self, value: Self):
        if not isinstance(value, type(self)):
            raise TypeError("Cannot copy from different type")
        for i in range(self.size()):
            self[i] = value[i]

    def _copy_(self) -> Self:
        if ctx():
            place = ctx().alloc(size=self._size_())
            result: Self = self._from_place_(place)
            result._copy_from_(self)
            return result
        else:
            assert isinstance(self._value, list)
            return self._with_value([value._copy_() for value in self._value])

    @classmethod
    def _alloc_(cls) -> Self:
        if ctx():
            place = ctx().alloc(size=cls._size_())
            return cls._from_place_(place)
        else:
            return cls._with_value([cls.element_type()._alloc_() for _ in range(cls.size())])

    @classmethod
    def _zero_(cls) -> Self:
        if ctx():
            place = ctx().alloc(size=cls._size_())
            result: Self = cls._from_place_(place)
            ctx().add_statements(*[IRSet(place.add_offset(i), IRConst(0)) for i in range(cls._size_())])
            return result
        else:
            return cls._with_value([cls.element_type()._zero_() for _ in range(cls.size())])

    def __len__(self):
        return self.size()

    @meta_fn
    def __getitem__(self, index: Num) -> T:
        index: Num = Num._accept_(get_positive_index(index, self.size()))
        if index._is_py_() and 0 <= index._as_py_() < self.size():
            const_index = index._as_py_()
            if isinstance(const_index, float) and not const_index.is_integer():
                raise ValueError("Array index must be an integer")
            const_index = int(const_index)
            if isinstance(self._value, list):
                if ctx():
                    return self._value[const_index]._get_()
                else:
                    return self._value[const_index]._get_()._as_py_()
            elif isinstance(self._value, BlockPlace):
                return (
                    self.element_type()
                    ._from_place_(self._value.add_offset(const_index * self.element_type()._size_()))
                    ._get_()
                )
            elif callable(self._value):
                return self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + Num(const_index * self.element_type()._size_())).ir())
                )
            else:
                raise InternalError("Unexpected array value")
        else:
            if not ctx():
                raise InternalError("Unexpected non-constant index")
            if isinstance(self._value, list | BlockPlace):
                base = ctx().rom[tuple(self._to_list_())] if isinstance(self._value, list) else self._value
                place = BlockPlace(
                    block=base.block,
                    index=(Num(base.index) + index * self.element_type()._size_()).index(),
                    offset=base.offset,
                )
                return self.element_type()._from_place_(place)._get_()
            elif callable(self._value):
                base_offset = index * Num(self.element_type()._size_())
                return self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + base_offset).ir())
                )
            else:
                raise InternalError("Unexpected array value")

    @meta_fn
    def __setitem__(self, index: Num, value: T):
        index: Num = Num._accept_(get_positive_index(index, self.size()))
        value = self.element_type()._accept_(value)
        if ctx():
            if isinstance(self._value, list):
                raise ValueError("Cannot mutate a compile time constant array")
            elif isinstance(self._value, BlockPlace):
                base = self._value
                place = (
                    base.add_offset(int(index._as_py_()) * self.element_type()._size_())
                    if index._is_py_()
                    else BlockPlace(
                        block=base.block,
                        index=(Num(base.index) + index * self.element_type()._size_()).index(),
                        offset=base.offset,
                    )
                )
                dst = self.element_type()._from_place_(place)
            elif callable(self._value):
                base_offset = index * Num(self.element_type()._size_())
                dst = self.element_type()._from_backing_source_(
                    lambda offset: self._value((Num(offset) + base_offset).ir())
                )
            else:
                raise InternalError("Unexpected array value")
            if self.element_type()._is_value_type_():
                dst._set_(value)
            else:
                dst._copy_from_(value)
        else:
            if not isinstance(self._value, list):
                raise InternalError("Unexpected mutation of non compile time constant array")
            const_index = index._as_py_()
            if isinstance(const_index, float) and not const_index.is_integer():
                raise ValueError("Array index must be an integer")
            const_index = int(const_index)
            if not 0 <= const_index < self.size():
                raise IndexError("Array index out of range")
            dst = self._value[const_index]
            if self.element_type()._is_value_type_():
                dst._set_(value)
            else:
                dst._copy_from_(value)

    def __eq__(self, other):
        if not isinstance(other, ArrayLike):
            return False
        if len(self) != len(other):
            return False
        i = 0
        while i < self.size():
            if self[i] != other[i]:
                return False
            i += 1
        return True

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(tuple(self[i] for i in range(self.size())))

    def __str__(self):
        if isinstance(self._value, BlockPlace) or callable(self._value):
            return f"{type(self).__name__}({self._value}...)"
        else:
            return f"{type(self).__name__}({', '.join(str(self[i]) for i in range(self.size()))})"

    def __repr__(self):
        if isinstance(self._value, BlockPlace) or callable(self._value):
            return f"{type(self).__name__}({self._value}...)"
        else:
            return f"{type(self).__name__}({', '.join(repr(self[i]) for i in range(self.size()))})"

    @meta_fn
    def __pos__(self) -> Self:
        """Return a copy of the array."""
        return self._copy_()



from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import Any, Self

from sonolus.backend.ir import IRConst, IRExpr, IRStmt
from sonolus.backend.place import BlockPlace


class BackingValue:
    def read(self) -> IRExpr:
        raise NotImplementedError()

    def write(self, value: IRExpr) -> IRStmt:
        raise NotImplementedError()


class ExprBackingValue(BackingValue):
    """A backing value that is backed by an expression."""

    def __init__(self, expr: IRExpr):
        self._expr = expr

    def read(self) -> IRExpr:
        return self._expr

    def write(self, value: IRExpr) -> IRStmt:
        raise RuntimeError("Value is read-only, cannot write to it")


type DataValue = BlockPlace | BackingValue | float | int | bool
type BackingSource = Callable[[IRExpr], BackingValue]


class Value:
    """Base class for values."""

    @classmethod
    def _is_concrete_(cls) -> bool:
        """Returns whether this type is concrete (i.e. can be instantiated)."""
        return False

    @classmethod
    @abstractmethod
    def _size_(cls) -> int:
        """Returns the size of this value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _is_value_type_(cls) -> bool:
        """Returns whether this is a value type.

        If this is true, the value behaves immutably and _set_ is supported.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        """Creates a value from a place."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _accepts_(cls, value: Any) -> bool:
        """Returns whether this value can accept the given value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _accept_(cls, value: Any) -> Self:
        """Accepts the given value."""
        raise NotImplementedError

    @abstractmethod
    def _is_py_(self) -> bool:
        """Returns whether this value is a valid Python value."""
        raise NotImplementedError

    @abstractmethod
    def _as_py_(self) -> Any:
        """Returns the Python equivalent of this value.

        Will fail if _is_py_ returns false.
        """
        raise NotImplementedError

    @classmethod
    def _from_backing_source_(cls, source: BackingSource) -> Self:
        """Creates a value from a backing source."""
        return cls._from_list_(source(IRConst(i)) for i in range(cls._size_()))

    @classmethod
    @abstractmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        """Creates a value from a list of data values."""
        raise NotImplementedError

    @abstractmethod
    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        """Converts this value to a list of data values."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        """Returns the keys to a flat representation of this value."""
        raise NotImplementedError

    def _to_flat_dict_(self, prefix: str, level_refs: dict[Any, str] | None = None) -> dict[str, DataValue | str]:
        """Converts this value to a flat dictionary."""
        return dict(zip(self._flat_keys_(prefix), self._to_list_(level_refs), strict=False))

    @abstractmethod
    def _get_(self) -> Self:
        """Implements access to the value.

        This is used when accessing a value from a record or array (and when storing into a record).
        For value (immutable) types, this makes a copy to preserve the appearance of immutability.
        For mutable types, this returns self.

        For instance:
        ```
        class X(Record):
            v: Num

        a = 1
        b = X(a) # (1) _get_() is called on a
        c = b.v # (2) _get_() is called on the value for v

        # (1) prevents this from changing the value of a
        # (2) prevents this from changing the value of c
        # Thus, both calls to _get_() are necessary to ensure values behave immutably.
        b.v = 2
        ```
        """
        raise NotImplementedError

    @abstractmethod
    def _set_(self, value: Self):
        """Implements assignment (=).

        This is only supported by value types.
        This method must not change the active context like by branching.

        In some places, = might instead call assign_, such as when setting the value of a record field
        with a reference type.
        """
        raise NotImplementedError

    @abstractmethod
    def _copy_from_(self, value: Self):
        """Implements copy assignment (@=).

        This is only supported by mutable reference types.
        """
        raise NotImplementedError

    @abstractmethod
    def _copy_(self) -> Self:
        """Returns a deep copy of this value."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _alloc_(cls) -> Self:
        """Allocates a new value which may be uninitialized."""
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def _zero_(cls) -> Self:
        """Returns a zero-initialized value of this type."""
        raise NotImplementedError

    def __imatmul__(self, other):
        self._copy_from_(other)
        return self


Value.__imatmul__._meta_fn_ = True



from sonolus.backend.ir import IRConst, IRGet, IRInstr, IRPureInstr, IRSet, IRStmt
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass
from sonolus.backend.place import BlockPlace, SSAPlace, TempBlock


class InlineVars(CompilerPass):
    def run(self, entry: BasicBlock) -> BasicBlock:
        use_counts: dict[SSAPlace, int] = {}
        definitions: dict[SSAPlace, IRStmt] = {}

        for block in traverse_cfg_preorder(entry):
            for stmt in block.statements:
                self.count_uses(stmt, use_counts)
                if isinstance(stmt, IRSet) and isinstance(stmt.place, SSAPlace):
                    definitions[stmt.place] = stmt.value
            self.count_uses(block.test, use_counts)

        for p, defn in definitions.items():
            while True:
                if isinstance(defn, IRGet) and isinstance(defn.place, SSAPlace) and defn.place in definitions:
                    inside_defn = definitions[defn.place]
                    if not self.is_inlinable(inside_defn):
                        break
                    defn = inside_defn
                    continue
                inlinable_uses = self.get_inlinable_uses(defn, set())
                subs = {}
                for inside_p in inlinable_uses:
                    if inside_p not in definitions:
                        continue
                    inside_defn = definitions[inside_p]
                    if not self.is_inlinable(inside_defn):
                        continue
                    if (isinstance(inside_defn, IRGet) and isinstance(inside_defn.place, SSAPlace)) or use_counts[
                        inside_p
                    ] == 1:
                        subs[inside_p] = inside_defn
                if not subs:
                    break
                defn = self.substitute(defn, subs)
            definitions[p] = defn

        valid = {p for p in definitions if self.is_inlinable(definitions[p]) and use_counts.get(p, 0) <= 1}

        for block in traverse_cfg_preorder(entry):
            new_statements = []
            for stmt in [*block.statements, block.test]:
                inlinable_uses = self.get_inlinable_uses(stmt, set())
                subs = {}
                for p in inlinable_uses:
                    if p not in valid:
                        continue
                    definition = definitions[p]
                    subs[p] = definition

                if subs:
                    new_statements.append(self.substitute(stmt, subs))
                else:
                    new_statements.append(stmt)

            block.statements = new_statements[:-1]
            block.test = new_statements[-1]

        return entry

    def substitute(self, stmt, subs):
        match stmt:
            case IRConst():
                return stmt
            case IRInstr(op=op, args=args):
                return IRInstr(op=op, args=[self.substitute(arg, subs) for arg in args])
            case IRPureInstr(op=op, args=args):
                return IRPureInstr(op=op, args=[self.substitute(arg, subs) for arg in args])
            case IRGet(place=place):
                if place in subs:
                    return subs[place]
                return stmt
            case IRSet(place=place, value=value):
                return IRSet(place=place, value=self.substitute(value, subs))
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")

    def count_uses(self, stmt, use_counts):
        match stmt:
            case IRConst():
                pass
            case IRInstr(op=_, args=args) | IRPureInstr(op=_, args=args):
                for arg in args:
                    self.count_uses(arg, use_counts)
            case IRGet(place=place):
                self.count_uses(place, use_counts)
            case IRSet(place=place, value=value):
                if not isinstance(place, SSAPlace):  # We don't want to count the definition itself
                    self.count_uses(place, use_counts)
                self.count_uses(value, use_counts)
            case SSAPlace():
                use_counts[stmt] = use_counts.get(stmt, 0) + 1
            case BlockPlace(block=block, index=index, offset=_):
                self.count_uses(block, use_counts)
                self.count_uses(index, use_counts)
            case int() | float():
                pass
            case TempBlock():
                pass
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")
        return use_counts

    def get_inlinable_uses(self, stmt, uses):
        match stmt:
            case IRConst():
                pass
            case IRInstr(op=_, args=args) | IRPureInstr(op=_, args=args):
                for arg in args:
                    self.get_inlinable_uses(arg, uses)
            case IRGet(place=place):
                if isinstance(place, SSAPlace):
                    uses.add(place)
            case IRSet(place=_, value=value):
                self.get_inlinable_uses(value, uses)
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")
        return uses

    def is_inlinable(self, stmt):
        match stmt:
            case IRConst():
                return True
            case IRInstr(op=op, args=args) | IRPureInstr(op=op, args=args):
                return not op.side_effects and op.pure and all(self.is_inlinable(arg) for arg in args)
            case IRGet():
                return isinstance(stmt.place, SSAPlace)
            case IRSet():
                return False
            case _:
                raise TypeError(f"Unexpected statement: {stmt}")



# ruff: noqa
def sin(x: float) -> float:
    """Compute the sine of x.

    Args:
        x: The angle in radians.

    Returns:
        The sine of x.
    """
    ...

def cos(x: float) -> float:
    """Compute the cosine of x.

    Args:
        x: The angle in radians.

    Returns:
        The cosine of x.
    """
    ...

def tan(x: float) -> float:
    """Compute the tangent of x.

    Args:
        x: The angle in radians.

    Returns:
        The tangent of x.
    """
    ...

def asin(x: float) -> float:
    """Compute the arcsine of x.

    Args:
        x: A value between -1 and 1.

    Returns:
        The arcsine of x in radians.
    """
    ...

def acos(x: float) -> float:
    """Compute the arccosine of x.

    Args:
        x: A value between -1 and 1.

    Returns:
        The arccosine of x in radians.
    """
    ...

def atan(x: float) -> float:
    """Compute the arctangent of x.

    Args:
        x: A numeric value.

    Returns:
        The arctangent of x in radians.
    """
    ...

def atan2(y: float, x: float) -> float:
    """Compute the arctangent of y / x considering the quadrant.

    Args:
        y: The y-coordinate.
        x: The x-coordinate.

    Returns:
        The arctangent of y / x in radians.
    """
    ...

def sinh(x: float) -> float:
    """Compute the hyperbolic sine of x.

    Args:
        x: A numeric value.

    Returns:
        The hyperbolic sine of x.
    """
    ...

def cosh(x: float) -> float:
    """Compute the hyperbolic cosine of x.

    Args:
        x: A numeric value.

    Returns:
        The hyperbolic cosine of x.
    """
    ...

def tanh(x: float) -> float:
    """Compute the hyperbolic tangent of x.

    Args:
        x: A numeric value.

    Returns:
        The hyperbolic tangent of x.
    """
    ...

def floor(x: float) -> int:
    """Return the largest integer less than or equal to x.

    Args:
        x: A numeric value.

    Returns:
        The floor of x.
    """
    ...

def ceil(x: float) -> int:
    """Return the smallest integer greater than or equal to x.

    Args:
        x: A numeric value.

    Returns:
        The ceiling of x.
    """
    ...

def trunc(x: float) -> int:
    """Truncate x to the nearest integer towards zero.

    Args:
        x: A numeric value.

    Returns:
        The truncated integer value of x.
    """
    ...

def log(x: float, base: float = ...) -> float:
    """Compute the logarithm of x to the given base.

    Args:
        x: The number for which to compute the logarithm.
        base: The base of the logarithm. If omitted, returns the natural logarithm of x.

    Returns:
        The logarithm of x to the specified base.
    """
    ...



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import ClassVar, Self, TypedDict

from sonolus.backend.optimize import optimize
from sonolus.backend.optimize.passes import CompilerPass
from sonolus.script.archetype import ArchetypeSchema
from sonolus.script.engine import Engine
from sonolus.script.level import Level


class Project:
    """A Sonolus.py project.

    Args:
        engine: The engine of the project.
        levels: The levels of the project.
        resources: The path to the resources of the project.
    """

    def __init__(
        self,
        engine: Engine,
        levels: list[Level] | None = None,
        resources: PathLike | None = None,
    ):
        self.engine = engine
        self.levels = levels or []
        self.resources = Path(resources or "resources")

    def with_levels(self, levels: list[Level]) -> Self:
        """Create a new project with the specified levels.

        Args:
            levels: The levels of the project.

        Returns:
            The new project.
        """
        return Project(self.engine, levels, self.resources)

    def dev(self, build_dir: PathLike, port: int = 8080, config: BuildConfig | None = None):
        """Start a development server for the project.

        Args:
            build_dir: The path to the build directory.
            port: The port of the development server.
            config: The build configuration.
        """
        from sonolus.build.cli import build_collection, run_server

        build_collection(self, Path(build_dir), config)
        run_server(Path(build_dir) / "site", port=port)

    def build(self, build_dir: PathLike, config: BuildConfig | None = None):
        """Build the project.

        Args:
            build_dir: The path to the build directory.
            config: The build configuration.
        """
        from sonolus.build.cli import build_project

        config = config or BuildConfig()
        build_project(self, Path(build_dir), config)

    def schema(self) -> ProjectSchema:
        """Generate the schema of the project.

        Returns:
            The schema of the project.
        """
        from sonolus.build.project import get_project_schema

        return get_project_schema(self)


class ProjectSchema(TypedDict):
    archetypes: list[ArchetypeSchema]


@dataclass
class BuildConfig:
    """A configuration for building an engine package."""

    MINIMAL_PASSES: ClassVar[Sequence[CompilerPass]] = optimize.MINIMAL_PASSES
    """The minimal list of compiler passes."""

    FAST_PASSES: ClassVar[Sequence[CompilerPass]] = optimize.FAST_PASSES
    """The list of compiler passes for faster builds."""

    STANDARD_PASSES: ClassVar[Sequence[CompilerPass]] = optimize.STANDARD_PASSES
    """The standard list of compiler passes."""

    passes: Sequence[CompilerPass] = optimize.STANDARD_PASSES
    """The list of compiler passes to use."""

    build_play: bool = True
    """Whether to build the play package."""

    build_watch: bool = True
    """Whether to build the watch package."""

    build_preview: bool = True
    """Whether to build the preview package."""

    build_tutorial: bool = True
    """Whether to build the tutorial package."""



from sonolus.backend.ir import IRConst, IRGet, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.backend.optimize.flow import BasicBlock, traverse_cfg_preorder
from sonolus.backend.optimize.passes import CompilerPass


class CoalesceFlow(CompilerPass):
    def run(self, entry: BasicBlock) -> BasicBlock:
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block in processed:
                continue
            processed.add(block)
            for edge in block.outgoing:
                while True:
                    dst = edge.dst
                    if dst.phis or dst.statements or len(dst.outgoing) != 1 or dst is block:
                        break
                    next_dst = next(iter(dst.outgoing)).dst
                    if next_dst.phis:
                        break
                    dst.incoming.remove(edge)
                    if not dst.incoming:
                        for dst_edge in dst.outgoing:
                            dst_edge.dst.incoming.remove(dst_edge)
                        processed.add(dst)
                    edge.dst = next_dst
                    next_dst.incoming.add(edge)
                    if dst is edge.dst:
                        break
            default_edge = next((edge for edge in block.outgoing if edge.cond is None), None)
            if default_edge is not None:
                for edge in [*block.outgoing]:
                    if edge is default_edge:
                        continue
                    if edge.dst is default_edge.dst:
                        block.outgoing.remove(edge)
                        edge.dst.incoming.remove(edge)
            if len(block.outgoing) != 1:
                queue.extend(edge.dst for edge in block.outgoing)
                continue
            next_block = next(iter(block.outgoing)).dst
            if len(next_block.incoming) != 1:
                queue.append(next_block)
                if not block.statements and not block.phis and not next_block.phis:
                    for edge in block.incoming:
                        edge.dst = next_block
                        next_block.incoming.add(edge)
                    for edge in block.outgoing:  # There should be exactly one
                        next_block.incoming.remove(edge)
                    if block is entry:
                        entry = next_block
                continue
            for p, args in next_block.phis.items():
                if block not in args:
                    continue
                block.statements.append(IRSet(p, IRGet(args[block])))
            block.statements.extend(next_block.statements)
            block.test = next_block.test
            block.outgoing = next_block.outgoing
            for edge in block.outgoing:
                edge.src = block
                dst = edge.dst
                for args in dst.phis.values():
                    if next_block in args:
                        args[block] = args.pop(next_block)
            processed.add(next_block)
            queue.extend(edge.dst for edge in block.outgoing)
            processed.remove(block)
            queue.append(block)
        return entry


class RewriteToSwitch(CompilerPass):
    """Rewrite if-else chains to switch statements.

    Note that this needs inlining (and dead code elimination) to be run first to really do anything useful.
    """

    def run(self, entry: BasicBlock) -> BasicBlock:
        self.ifs_to_switch(entry)
        self.combine_blocks(entry)
        self.remove_unreachable(entry)
        return entry

    def ifs_to_switch(self, entry: BasicBlock):
        for block in traverse_cfg_preorder(entry):
            if len(block.outgoing) != 2 or {edge.cond for edge in block.outgoing} != {None, 0}:
                continue
            test = block.test
            if not isinstance(test, IRPureInstr) or test.op != Op.Equal:
                continue
            assert len(test.args) == 2
            if isinstance(test.args[0], IRConst):
                const, other = test.args
            elif isinstance(test.args[1], IRConst):
                other, const = test.args
            else:
                continue
            block.test = other
            for edge in block.outgoing:
                if edge.cond is None:
                    edge.cond = const.value
                else:
                    edge.cond = None

    def combine_blocks(self, entry: BasicBlock):
        queue = [entry]
        processed = set()
        while queue:
            block = queue.pop()
            if block in processed:
                continue
            processed.add(block)
            queue.extend(edge.dst for edge in block.outgoing)

            default_edge = next((edge for edge in block.outgoing if edge.cond is None), None)
            if default_edge is None:
                continue

            next_block = default_edge.dst
            if (
                len(next_block.incoming) > 1
                or next_block.statements
                or next_block.phis
                or block.test != next_block.test
                or block is next_block
                or next_block is entry
            ):
                continue

            outgoing_by_cond = {edge.cond: edge for edge in block.outgoing}
            assert len(outgoing_by_cond) == len(block.outgoing)
            outgoing_by_cond.pop(None)
            for edge in next_block.outgoing:
                if edge.cond in outgoing_by_cond:
                    # This edge is unreachable since an equivalent edge would have been taken
                    edge.dst.incoming.remove(edge)
                    continue
                outgoing_by_cond[edge.cond] = edge
                edge.src = block
                for args in edge.dst.phis.values():
                    if next_block in args:
                        args[block] = args.pop(next_block)
            block.outgoing = set(outgoing_by_cond.values())
            processed.add(next_block)
            queue.append(block)
            processed.remove(block)

    def remove_unreachable(self, entry: BasicBlock):
        reachable = {*traverse_cfg_preorder(entry)}
        for block in traverse_cfg_preorder(entry):
            block.incoming = {edge for edge in block.incoming if edge.src in reachable}
            block.outgoing = {edge for edge in block.outgoing if edge.dst in reachable}


class NormalizeSwitch(CompilerPass):
    """Normalize branches like cond -> case a, case a + b, case a + 2b to ((cond - a) / b) -> case 0, case 1, case 2."""

    def run(self, entry: BasicBlock) -> BasicBlock:
        for block in traverse_cfg_preorder(entry):
            cases = {edge.cond for edge in block.outgoing}
            if len(cases) <= 2:
                continue
            assert None in cases, "Non-terminal blocks should always have a default edge"
            cases.remove(None)
            offset, stride = self.get_offset_stride(cases)
            if offset is None or (offset == 0 and stride == 1):
                continue
            for edge in block.outgoing:
                if edge.cond is None:
                    continue
                edge.cond = (edge.cond - offset) // stride
            if offset != 0:
                block.test = IRPureInstr(Op.Subtract, [block.test, IRConst(offset)])
            if stride != 1:
                block.test = IRPureInstr(Op.Divide, [block.test, IRConst(stride)])
        return entry

    def get_offset_stride(self, cases: set[int]) -> tuple[int | None, int | None]:
        cases = sorted(cases)
        offset = cases[0]
        stride = cases[1] - offset
        if int(offset) != offset or int(stride) != stride:
            return None, None
        for i, case in enumerate(cases[2:], 2):
            if case != offset + i * stride:
                return None, None
        return offset, stride



from __future__ import annotations

import typing
from enum import Enum
from types import UnionType
from typing import Annotated, Any, ClassVar, Literal, Self, TypeVar, get_origin

from sonolus.script.internal.impl import meta_fn, validate_value
from sonolus.script.internal.value import Value

type AnyType = type[Value] | PartialGeneric | TypeVar


def validate_type_arg(arg: Any) -> Any:
    arg = validate_value(arg)
    if not arg._is_py_():
        raise TypeError(f"Expected a compile-time constant type argument, got {arg}")
    result = arg._as_py_()
    if hasattr(result, "_type_mapping_"):
        return result._type_mapping_
    if get_origin(result) is Annotated:
        return result.__args__[0]
    if get_origin(result) is Literal:
        return result.__args__[0]
    return result


def validate_type_spec(spec: Any) -> PartialGeneric | TypeVar | type[Value]:
    spec = validate_type_arg(spec)
    if isinstance(spec, type) and issubclass(spec, Enum):
        # For values like IntEnum subclasses, this will call validate_type_spec(IntEnum),
        # which in turn will call it on int, so this works.
        spec = validate_type_spec(spec.__mro__[1])
    if isinstance(spec, PartialGeneric | TypeVar) or (isinstance(spec, type) and issubclass(spec, Value)):
        return spec
    if typing.get_origin(spec) is UnionType:
        args = typing.get_args(spec)
        validated_args = {validate_type_arg(arg) for arg in args}
        if len(validated_args) == 1:
            return validated_args.pop()
    raise TypeError(f"Invalid type spec: {spec}")


def validate_concrete_type(spec: Any) -> type[Value]:
    spec = validate_type_spec(spec)
    if isinstance(spec, type) and issubclass(spec, Value) and spec._is_concrete_():
        return spec
    raise TypeError(f"Expected a concrete type, got {spec}")


def validate_type_args(args) -> tuple[Any, ...]:
    if not isinstance(args, tuple):
        args = (args,)
    return tuple(validate_type_arg(arg) for arg in args)


def contains_incomplete_type(args) -> bool:
    if not isinstance(args, tuple):
        args = (args,)
    return any(isinstance(arg, TypeVar | PartialGeneric) for arg in args)


def format_type_arg(arg: Any) -> str:
    if isinstance(arg, type):
        return arg.__name__
    return f"{arg}"


class GenericValue(Value):
    _parameterized_: ClassVar[dict[tuple[Any, ...], type[Self]]] = {}
    _type_args_: ClassVar[tuple[Any, ...] | None] = None
    _type_vars_to_args_: ClassVar[dict[TypeVar, Any] | None] = None

    def __init__(self):
        if self._type_args_ is None:
            raise TypeError(f"Missing type arguments for {self.__class__.__name__}")

    @classmethod
    def _validate_type_args_(cls, args: tuple[Any, ...]) -> tuple[Any, ...]:
        """Validate the type arguments and return them as a tuple.

        This may be called with PartialGeneric or TypeVar instances inside args.
        """
        if len(args) != len(cls.__type_params__):
            raise TypeError(f"Expected {len(cls.__type_params__)} type arguments, got {len(args)}")
        return args

    @classmethod
    def _is_concrete_(cls) -> bool:
        return cls._type_args_ is not None

    @classmethod
    @meta_fn
    def type_var_value(cls, var: TypeVar) -> Any:
        if isinstance(var, Value):
            var = var._as_py_()
        if cls._type_args_ is None:
            raise TypeError(f"Type {cls.__name__} is not parameterized")
        if var in cls._type_vars_to_args_:
            return cls._type_vars_to_args_[var]
        raise TypeError(f"Missing type argument for {var}")

    def __class_getitem__(cls, args: Any) -> type[Self]:
        if cls._type_args_ is not None:
            raise TypeError(f"Type {cls.__name__} is already parameterized")
        args = validate_type_args(args)
        args = cls._validate_type_args_(args)
        if contains_incomplete_type(args):
            return PartialGeneric(cls, args)
        if args not in cls._parameterized_:
            cls._parameterized_[args] = cls._get_parameterized(args)
        return cls._parameterized_[args]

    @classmethod
    def _get_parameterized(cls, args: tuple[Any, ...]) -> type[Self]:
        class Parameterized(cls):
            _type_args_ = args
            _type_vars_to_args_ = dict(zip(cls.__type_params__, args, strict=True))  # noqa: RUF012

        if args:
            Parameterized.__name__ = f"{cls.__name__}[{', '.join(format_type_arg(arg) for arg in args)}]"
            Parameterized.__qualname__ = f"{cls.__qualname__}[{', '.join(format_type_arg(arg) for arg in args)}]"
        else:
            Parameterized.__name__ = cls.__name__
            Parameterized.__qualname__ = cls.__qualname__
        Parameterized.__module__ = cls.__module__
        return Parameterized


class PartialGeneric[T: GenericValue]:
    def __init__(self, base: type[T], args: tuple[Any, ...]):
        self.base = base
        self.args = args

    def __repr__(self):
        return f"PartialGeneric({self.base!r}, {self.args!r})"

    def __str__(self):
        params = ", ".join(format_type_arg(arg) for arg in self.args)
        return f"{self.base.__name__}?[{params}]"

    def __call__(self, *args, **kwargs):
        instance = self.base(*args, **kwargs)
        # Throw an error if it fails
        accept_and_infer_types(self, instance, {})
        return instance


def infer_and_validate_types(dst: Any, src: Any, results: dict[TypeVar, Any] | None = None) -> dict[TypeVar, Any]:
    results = results if results is not None else {}
    match dst:
        case TypeVar():
            if dst in results:
                if results[dst] != src:
                    raise TypeError(
                        f"Conflicting types for {dst}: {format_type_arg(results[dst])} and {format_type_arg(src)}"
                    )
            else:
                results[dst] = validate_type_arg(src)
        case PartialGeneric():
            if not isinstance(src, type) or not issubclass(src, dst.base):
                raise TypeError(f"Expected type {dst.base.__name__}, got {format_type_arg(src)}")
            assert issubclass(src, GenericValue)
            for d, s in zip(dst.args, src._type_args_, strict=True):
                infer_and_validate_types(d, s, results)
        case _:
            if (
                src != dst  # noqa: PLR1714
                and dst != Any
                and not (isinstance(dst, type) and isinstance(src, type) and issubclass(src, dst))
            ):
                raise TypeError(f"Expected {format_type_arg(dst)}, got {format_type_arg(src)}")
    return results


def accept_and_infer_types(dst: Any, val: Any, results: dict[TypeVar, Any]) -> Value:
    val = validate_value(val)
    match dst:
        case TypeVar():
            infer_and_validate_types(dst, type(val), results)
            return val
        case PartialGeneric():
            val = dst.base._accept_(val)
            infer_and_validate_types(dst, type(val), results)
            return val
        case type():
            return dst._accept_(val)
        case _:
            raise TypeError(f"Expected a type, got {format_type_arg(dst)}")


def validate_and_resolve_type(dst: Any, results: dict[TypeVar, Any]) -> Any:
    match dst:
        case TypeVar():
            if dst in results:
                return results[dst]
            raise TypeError(f"Missing type for {dst}")
        case PartialGeneric():
            return dst.base[tuple(validate_and_resolve_type(arg, results) for arg in dst.args)]
        case _:
            return dst



from collections.abc import Iterable
from typing import Any, Self

from sonolus.backend.place import BlockPlace
from sonolus.script.internal.value import DataValue, Value


class TransientValue(Value):
    @classmethod
    def _is_concrete_(cls) -> bool:
        return True

    @classmethod
    def _size_(cls) -> int:
        return 0

    @classmethod
    def _is_value_type_(cls) -> bool:
        return False

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        raise TypeError(f"{cls.__name__} cannot be dereferenced")

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        raise TypeError(f"{cls.__name__} cannot be constructed from list")

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue | str]:
        raise TypeError(f"{type(self).__name__} cannot be deconstructed to list")

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        raise TypeError(f"{cls.__name__} cannot be flattened")

    def _get_(self) -> Self:
        return self

    def _set_(self, value: Self) -> None:
        if value is not self:
            raise TypeError(f"{type(self).__name__} is immutable")

    def _copy_from_(self, value: Self):
        raise TypeError(f"{type(self).__name__} is immutable")

    def _copy_(self) -> Self:
        raise TypeError(f"{type(self).__name__} cannot be copied")

    @classmethod
    def _alloc_(cls) -> Self:
        raise TypeError(f"{cls.__name__} is not allocatable")

    @classmethod
    def _zero_(cls) -> Self:
        raise TypeError(f"{cls.__name__} does not have a zero value")



from __future__ import annotations

import json
from collections.abc import Iterator
from os import PathLike
from pathlib import Path
from typing import Any

from sonolus.build.collection import Asset, load_asset
from sonolus.script.archetype import PlayArchetype, StandardArchetypeName, StandardImport
from sonolus.script.metadata import AnyText, Tag, as_localization_text


class ExportedLevel:
    """An exported Sonolus level."""

    def __init__(
        self,
        *,
        item: dict,
        cover: bytes,
        bgm: bytes,
        preview: bytes | None,
        data: bytes,
    ):
        self.item = item
        self.cover = cover
        self.bgm = bgm
        self.preview = preview
        self.data = data

    def write_to_dir(self, path: PathLike):
        """Write the exported level to a directory."""
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        (path / "item.json").write_text(json.dumps(self.item, ensure_ascii=False), encoding="utf-8")
        (path / "cover").write_bytes(self.cover)
        (path / "bgm").write_bytes(self.bgm)
        if self.preview is not None:
            (path / "preview").write_bytes(self.preview)
        (path / "data").write_bytes(self.data)


class Level:
    """A Sonolus level.

    Args:
        name: The name of the level.
        title: The title of the level.
        rating: The rating of the level.
        artists: The artists of the level.
        author: The author of the level.
        cover: The cover of the level.
        bgm: The background music of the level.
        data: The data of the level.
        use_skin: The skin to use, overriding the engine skin.
        use_background: The background to use, overriding the engine background.
        use_effect: The effect to use, overriding the engine effect.
        use_particle: The particle to use, overriding the engine particle.
        tags: The tags of the level.
        description: The description of the level.
        meta: Additional metadata of the level.
    """

    version = 1

    def __init__(
        self,
        *,
        name: str,
        title: AnyText | None = None,
        rating: int = 0,
        artists: AnyText = "Unknown",
        author: AnyText = "Unknown",
        cover: Asset | None = None,
        bgm: Asset | None = None,
        preview: Asset | None = None,
        data: LevelData,
        use_skin: str | None = None,
        use_background: str | None = None,
        use_effect: str | None = None,
        use_particle: str | None = None,
        tags: list[Tag] | None = None,
        description: AnyText | None = None,
        meta: Any = None,
    ) -> None:
        self.name = name
        self.title = as_localization_text(title or name)
        self.rating = rating
        self.artists = as_localization_text(artists)
        self.author = as_localization_text(author)
        self.cover = cover
        self.bgm = bgm
        self.preview = preview
        self.data = data
        self.use_skin = use_skin
        self.use_background = use_background
        self.use_effect = use_effect
        self.use_particle = use_particle
        self.tags = tags or []
        self.description = as_localization_text(description) if description is not None else None
        self.meta = meta

    def export(self, engine_name: str) -> ExportedLevel:
        """Export the level in a sonolus-pack compatible format.

        Args:
            engine_name: The name of the engine this level is for.

        Returns:
            The exported level.
        """
        from sonolus.build.level import package_level_data
        from sonolus.build.project import BLANK_AUDIO, BLANK_PNG

        item = {
            "version": self.version,
            "rating": self.rating,
            "engine": engine_name,
            "useSkin": {"useDefault": True} if self.use_skin is None else {"useDefault": False, "item": self.use_skin},
            "useBackground": {"useDefault": True}
            if self.use_background is None
            else {"useDefault": False, "item": self.use_background},
            "useEffect": {"useDefault": True}
            if self.use_effect is None
            else {"useDefault": False, "item": self.use_effect},
            "useParticle": {"useDefault": True}
            if self.use_particle is None
            else {"useDefault": False, "item": self.use_particle},
            "title": self.title,
            "artists": self.artists,
            "author": self.author,
            "tags": [tag.as_dict() for tag in self.tags],
        }
        if self.description is not None:
            item["description"] = self.description
        if self.meta is not None:
            item["meta"] = self.meta
        return ExportedLevel(
            item=item,
            cover=load_asset(self.cover) if self.cover is not None else BLANK_PNG,
            bgm=load_asset(self.bgm) if self.bgm is not None else BLANK_AUDIO,
            preview=load_asset(self.preview) if self.preview is not None else None,
            data=package_level_data(self.data),
        )


type EntityListArg = list[PlayArchetype | EntityListArg]


def flatten_entities(entities: EntityListArg) -> Iterator[PlayArchetype]:
    """Flatten a list of entities.

    Args:
        entities: The list of entities.

    Yields:
        The flattened entities.
    """
    if isinstance(entities, list):
        for entity in entities:
            yield from flatten_entities(entity)
    else:
        yield entities


class LevelData:
    """The data of a Sonolus level.

    Args:
        bgm_offset: The background music audio offset.
        entities: The entities of the level.
    """

    bgm_offset: float
    entities: list[PlayArchetype]

    def __init__(self, bgm_offset: float, entities: list[PlayArchetype]) -> None:
        self.bgm_offset = bgm_offset
        self.entities = [*flatten_entities(entities)]


class BpmChange(PlayArchetype):
    """The standard bpm change archetype."""

    name = StandardArchetypeName.BPM_CHANGE

    beat: StandardImport.BEAT
    bpm: StandardImport.BPM


class TimescaleChange(PlayArchetype):
    """The standard timescale change archetype."""

    name = StandardArchetypeName.TIMESCALE_CHANGE

    beat: StandardImport.BEAT
    timescale: StandardImport.TIMESCALE



from __future__ import annotations

import operator
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, Self, TypeGuard, final, runtime_checkable

from sonolus.backend.ir import IRConst, IRExpr, IRGet, IRPureInstr, IRSet
from sonolus.backend.ops import Op
from sonolus.backend.place import BlockPlace
from sonolus.script.internal.context import ctx
from sonolus.script.internal.error import InternalError
from sonolus.script.internal.impl import meta_fn
from sonolus.script.internal.value import BackingValue, DataValue, ExprBackingValue, Value


class _NumMeta(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, float | int | bool) or _is_num(instance)


def _is_num(value: Any) -> TypeGuard[Num]:
    """Check if a value is a precisely Num instance."""
    return type.__instancecheck__(Num, value)  # type: ignore # noqa: PLC2801


@final
class _Num(Value, metaclass=_NumMeta):
    # This works for ints, floats, and bools
    # Since we don't support complex numbers, real is equal to the original number
    __match_args__ = ("real",)

    data: DataValue

    def __init__(self, data: DataValue | IRExpr):
        if isinstance(data, complex):
            raise TypeError("Cannot create a Num from a complex number")
        if isinstance(data, int):
            data = float(data)
        if isinstance(data, IRConst | IRPureInstr | IRGet):
            data = ExprBackingValue(data)
        if _is_num(data):
            raise InternalError("Cannot create a Num from a Num")
        if not isinstance(data, BlockPlace | BackingValue | float | int | bool):
            raise TypeError(f"Cannot create a Num from {type(data)}")
        self.data = data

    def __str__(self) -> str:
        if isinstance(self.data, float) and self.data.is_integer():
            return str(int(self.data))
        return str(self.data)

    def __repr__(self) -> str:
        return f"Num({self.data})"

    @classmethod
    def _is_concrete_(cls) -> bool:
        return True

    @classmethod
    def _size_(cls) -> int:
        return 1

    @classmethod
    def _is_value_type_(cls) -> bool:
        return True

    @classmethod
    def _from_place_(cls, place: BlockPlace) -> Self:
        return cls(place)

    @classmethod
    def _accepts_(cls, value: Value) -> bool:
        return isinstance(value, Num | float | int | bool)

    @classmethod
    def _accept_(cls, value: Any) -> Self:
        if not cls._accepts_(value):
            raise TypeError(f"Cannot accept {value}")
        if _is_num(value):
            return value
        return cls(value)

    def _is_rom_constant(self) -> bool:
        return (
            ctx()
            and isinstance(self.data, BlockPlace)
            and self.data.block == ctx().blocks.EngineRom
            and isinstance(self.data.index, int)
        )

    def _is_py_(self) -> bool:
        return isinstance(self.data, float | int | bool) or self._is_rom_constant()

    def _as_py_(self) -> Any:
        if not self._is_py_():
            raise ValueError("Not a compile time constant Num")
        if self._is_rom_constant():
            return ctx().rom.get_value(self.data.index + self.data.offset)
        if self.data.is_integer():
            return int(self.data)
        return self.data

    @classmethod
    def _from_list_(cls, values: Iterable[DataValue]) -> Self:
        value = next(iter(values))
        return Num(value)

    def _to_list_(self, level_refs: dict[Any, str] | None = None) -> list[DataValue]:
        return [self.data]

    @classmethod
    def _flat_keys_(cls, prefix: str) -> list[str]:
        return [prefix]

    def _get_(self) -> Self:
        if ctx():
            place = ctx().alloc(size=1)
            if isinstance(self.data, BlockPlace):
                ctx().check_readable(self.data)
            ctx().add_statements(IRSet(place, self.ir()))
            return Num(place)
        else:
            return Num(self.data)

    def _set_(self, value: Self):
        value = Num._accept_(value)
        if ctx():
            match self.data:
                case BackingValue():
                    self.data.write(value.ir())
                case BlockPlace():
                    ctx().check_writable(self.data)
                    ctx().add_statements(IRSet(self.data, value.ir()))
                case _:
                    raise ValueError("Cannot set a read-only value")
        else:
            self.data = value.data

    def _copy_from_(self, value: Self):
        raise ValueError("Cannot assign to a number")

    def _copy_(self) -> Self:
        if ctx():
            return self._get_()
        else:
            return Num(self.data)

    @classmethod
    def _alloc_(cls) -> Self:
        if ctx():
            return Num(ctx().alloc(size=1))
        else:
            return Num(-1)

    @classmethod
    def _zero_(cls) -> Self:
        if ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRConst(0)))
            return cls(result_place)
        else:
            return cls(0)

    def ir(self):
        match self.data:
            case BlockPlace():
                return IRGet(self.data)
            case BackingValue():
                return self.data.read()
            case _:
                return IRConst(self.data)

    def index(self) -> int | BlockPlace:
        if isinstance(self.data, BlockPlace):
            return self._get_().data
        return self.data

    def _bin_op(self, other: Self, const_fn: Callable[[Self, Self], Self | None], ir_op: Op) -> Self:
        if not Num._accepts_(other):
            return NotImplemented
        other = Num._accept_(other)
        const_value = const_fn(self, other)
        if const_value is not None:
            return const_value
        if ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRPureInstr(ir_op, [self.ir(), other.ir()])))
            return Num(result_place)
        else:
            raise InternalError("Unexpected call on non-comptime Num instance outside a compilation context")

    def _unary_op(self, py_fn: Callable[[float], float], ir_op: Op) -> Self:
        if self._is_py_():
            return Num(py_fn(self._as_py_()))
        elif ctx():
            result_place = ctx().alloc(size=1)
            ctx().add_statements(IRSet(result_place, IRPureInstr(ir_op, [self.ir()])))
            return Num(result_place)
        else:
            raise InternalError("Unexpected call on non-comptime Num instance outside a compilation context")

    @meta_fn
    def __eq__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() == b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.Equal)

    def __hash__(self):
        if self._is_py_():
            return hash(self._as_py_())
        raise ValueError("Cannot hash non compile time constant Num")

    @meta_fn
    def __ne__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() != b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.NotEqual)

    @meta_fn
    def __lt__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() < b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.Less)

    @meta_fn
    def __le__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() <= b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.LessOr)

    @meta_fn
    def __gt__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() > b._as_py_())
            if a.data == b.data:
                return Num(False)
            return None

        return self._bin_op(other, const_fn, Op.Greater)

    @meta_fn
    def __ge__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() >= b._as_py_())
            if a.data == b.data:
                return Num(True)
            return None

        return self._bin_op(other, const_fn, Op.GreaterOr)

    @meta_fn
    def __add__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() + b._as_py_())
            if a._is_py_() and a._as_py_() == 0:
                return b
            if b._is_py_() and b._as_py_() == 0:
                return a
            return None

        return self._bin_op(other, const_fn, Op.Add)

    @meta_fn
    def __sub__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() - b._as_py_())
            if a._is_py_() and a._as_py_() == 0:
                return -b
            if b._is_py_() and b._as_py_() == 0:
                return a
            return None

        return self._bin_op(other, const_fn, Op.Subtract)

    @meta_fn
    def __mul__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() * b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return Num(0)
                if a._as_py_() == 1:
                    return b
            if b._is_py_():
                if b._as_py_() == 0:
                    return Num(0)
                if b._as_py_() == 1:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Multiply)

    @meta_fn
    def __truediv__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() / b._as_py_())
            if b._is_py_():
                if b._as_py_() == 1:
                    return a
                if b._as_py_() == -1:
                    return -a
            return None

        return self._bin_op(other, const_fn, Op.Divide)

    @meta_fn
    def __floordiv__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() // b._as_py_())
            if b._is_py_():
                if b._as_py_() == 1:
                    return a
                if b._as_py_() == -1:
                    return -a
            return None

        return self._bin_op(other, const_fn, Op.Divide)._unary_op(lambda x: x, Op.Floor)

    @meta_fn
    def __mod__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                if b._as_py_() == 0:
                    return None
                return Num(a._as_py_() % b._as_py_())
            return None

        return self._bin_op(other, const_fn, Op.Mod)

    @meta_fn
    def __pow__(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                try:
                    return Num(a._as_py_() ** b._as_py_())
                except OverflowError:
                    return None
            if b._is_py_():
                if b._as_py_() == 0:
                    return Num(1)
                if b._as_py_() == 1:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Power)

    @meta_fn
    def __neg__(self) -> Self:
        return self._unary_op(operator.neg, Op.Negate)

    @meta_fn
    def __pos__(self) -> Self:
        return self

    @meta_fn
    def __abs__(self) -> Self:
        return self._unary_op(abs, Op.Abs)

    @meta_fn
    def __bool__(self):
        if ctx():
            return self != 0
        else:
            if self._is_py_():
                return bool(self._as_py_())
            raise ValueError("Cannot convert non compile time constant Num to bool")

    def and_(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() and b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return a
                else:
                    return b
            return None

        return self._bin_op(other, const_fn, Op.And)

    def or_(self, other) -> Self:
        def const_fn(a: Self, b: Self) -> Num | None:
            if a._is_py_() and b._is_py_():
                return Num(a._as_py_() or b._as_py_())
            if a._is_py_():
                if a._as_py_() == 0:
                    return b
                else:
                    return a
            return None

        return self._bin_op(other, const_fn, Op.Or)

    def not_(self) -> Self:
        return self._unary_op(operator.not_, Op.Not)

    @property
    def real(self) -> Self:
        return self

    @property
    def imag(self) -> Self:
        return 0


if TYPE_CHECKING:
    from typing import Protocol

    @runtime_checkable
    class Num[T](Protocol, int, bool, float):
        def __add__(self, other: T, /) -> Num: ...
        def __sub__(self, other: T, /) -> Num: ...
        def __mul__(self, other: T, /) -> Num: ...
        def __truediv__(self, other: T, /) -> Num: ...
        def __floordiv__(self, other: T, /) -> Num: ...
        def __mod__(self, other: T, /) -> Num: ...
        def __pow__(self, other: T, /) -> Num: ...

        def __neg__(self, /) -> Num: ...
        def __pos__(self, /) -> Num: ...
        def __abs__(self, /) -> Num: ...

        def __eq__(self, other: Any, /) -> bool: ...
        def __ne__(self, other: Any, /) -> bool: ...
        def __lt__(self, other: T, /) -> bool: ...
        def __le__(self, other: T, /) -> bool: ...
        def __gt__(self, other: T, /) -> bool: ...
        def __ge__(self, other: T, /) -> bool: ...

        def __hash__(self, /) -> int: ...

        @property
        def real(self) -> Num: ...

        @property
        def imag(self) -> Num: ...
else:
    # Need to do this to satisfy type checkers (especially Pycharm)
    _Num.__name__ = "Num"
    _Num.__qualname__ = "Num"
    globals()["Num"] = _Num



class InternalError(RuntimeError):
    """Represents an error occurring due to a violation of an internal invariant.

    This indicates there is a bug in sonolus.py, or that internal details have been used incorrectly.
    """

    def __init__(self, message: str):
        super().__init__(message)


class CompilationError(RuntimeError):
    """Represents an error occurring during the compilation of a script."""

    def __init__(self, message: str):
        super().__init__(message)
